"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from datetime import timedelta
from datetime import timezone

from pytest import mark

from ..parse import parse_time
from ..parse import shift_time
from ..parse import since_time
from ..parse import string_time
from ..utils import utcdatetime



def test_parse_time() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    dtime = utcdatetime()
    delta = timedelta(seconds=1)


    parsed = parse_time(
        '1/1/1970 6:00am')

    assert parsed.year == 1970
    assert parsed.month == 1
    assert parsed.day == 1
    assert parsed.hour == 6


    parsed = parse_time(
        '12/31/1969 6:00pm',
        tzname='US/Central')

    assert parsed.year == 1969
    assert parsed.month == 12
    assert parsed.day == 31
    assert parsed.hour == 18


    parsed = parse_time(0)
    assert parsed.year == 1970

    parsed = parse_time('0')
    assert parsed.year == 1970


    parsed = parse_time('max')

    assert parsed.year == 9999
    assert parsed.month == 12
    assert parsed.day == 31
    assert parsed.hour == 23

    parsed = parse_time('min')

    assert parsed.year == 1
    assert parsed.month == 1
    assert parsed.day == 1
    assert parsed.hour == 0


    parsed = parse_time('now')
    assert parsed - dtime <= delta

    parsed = parse_time(None)
    assert parsed - dtime <= delta

    parsed = parse_time('None')
    assert parsed - dtime <= delta


    parsed = parse_time('+1y')
    assert parsed - dtime > delta


    _parsed = parse_time(parsed)
    assert _parsed == parsed



@mark.parametrize(
    'notate,expect',
    [('+1y', (1981, 1, 1)),
     ('-1y', (1979, 1, 1)),
     ('+1y@s', (1981, 1, 1)),
     ('-1y@s', (1979, 1, 1)),
     ('+1y@s+1h', (1981, 1, 1, 1)),
     ('-1y@s-1h', (1978, 12, 31, 23)),
     ('+1mon', (1980, 2, 1)),
     ('-1mon', (1979, 12, 1)),
     ('+1mon@m', (1980, 2, 1)),
     ('-1mon@m', (1979, 12, 1)),
     ('+1mon@m+1mon', (1980, 3, 1)),
     ('-1mon@m-1mon', (1979, 11, 1)),
     ('+1w', (1980, 1, 8)),
     ('-1w', (1979, 12, 25)),
     ('+1w@h', (1980, 1, 8)),
     ('-1w@h', (1979, 12, 25)),
     ('+1w@h+1w', (1980, 1, 15)),
     ('-1w@h-1w', (1979, 12, 18)),
     ('+1d', (1980, 1, 2)),
     ('-1d', (1979, 12, 31)),
     ('+1d@d', (1980, 1, 2)),
     ('-1d@d', (1979, 12, 31)),
     ('+1d@d+30d', (1980, 2, 1)),
     ('-1d@d-30d', (1979, 12, 1))])
def test_shift_time(
    notate: str,
    expect: tuple[int, ...],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param notate: Syntax compatable using snaptime library.
    :param expect: Expected output from the testing routine.
    """

    anchor = utcdatetime(1980, 1, 1)
    dtime = utcdatetime(*expect)

    parsed = shift_time(notate, anchor)

    assert parsed == dtime



def test_string_time() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    utc = timezone.utc
    expect = (
        utcdatetime(1980, 1, 1)
        .astimezone(None))


    strings = [
        '1980-01-01T00:00:00Z',
        '1980-01-01T00:00:00 +0000',
        '1980-01-01T00:00:00',
        '1980-01-01 00:00:00 +0000',
        '1980-01-01 00:00:00']

    for string in strings:

        parsed = (
            string_time(string)
            .astimezone(utc))

        assert parsed == expect


    parsed = string_time(
        '1980_01_01',
        formats=['%Y_%m_%d'])

    assert parsed == expect


    parsed = string_time(
        '1980_01_01',
        formats='%Y_%m_%d')

    assert parsed == expect


    parsed = (
        string_time(
            '1979-12-31 18:00:00',
            tzname='US/Central'))

    assert parsed == expect



def test_since_time() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    parsed = shift_time('-1s')
    since = since_time(parsed)

    assert since >= 1
    assert since < 2


    parsed = shift_time('+1s')
    since = since_time(parsed)

    assert since > 0
    assert since < 2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from .common import PARSABLE
from .common import SCHEDULE
from .time import Time
from ..types import BaseModel
from ..types import DictStrAny



_TIMERS = dict[str, 'TimerParams']
_WINDOWS = dict[str, 'WindowParams']



class TimerParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    timer: Annotated[
        float,
        Field(...,
              description='Seconds used for the interval')]

    start: Annotated[
        str,
        Field(...,
              description='Optional value of timer start',
              min_length=1)]


    def __init__(
        self,
        timer: int | float,
        start: PARSABLE = 'now',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        data: DictStrAny = {}


        if timer is not None:
            timer = float(timer)

        if start is not None:
            start = Time(start)


        data['timer'] = timer

        if start is not None:
            data['start'] = start.subsec


        super().__init__(**data)



class TimersParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    timers: Annotated[
        _TIMERS,
        Field(...,
              description='Seconds used for the interval',
              min_length=0)]


    def __init__(
        self,
        timers: Optional[_TIMERS] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        timers = timers or {}

        super().__init__(**{
            'timers': timers})



class WindowParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    window: Annotated[
        SCHEDULE,
        Field(...,
              description='Period for scheduling window',
              min_length=1)]

    start: Annotated[
        Optional[str],
        Field(None,
              description='Determine when scope begins',
              min_length=1)]

    stop: Annotated[
        Optional[str],
        Field(None,
              description='Determine when scope ends',
              min_length=1)]

    anchor: Annotated[
        Optional[str],
        Field(None,
              description='Optional anchor of the window',
              min_length=1)]

    delay: Annotated[
        float,
        Field(...,
              description='Time period of schedule delay',
              ge=0)]


    def __init__(
        self,
        window: SCHEDULE | int,
        *,
        start: Optional[PARSABLE] = None,
        stop: Optional[PARSABLE] = None,
        anchor: Optional[PARSABLE] = None,
        delay: int | float = 0.0,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        data: DictStrAny = {}


        if isinstance(window, int):
            window = {'seconds': window}

        if start is not None:
            start = Time(start)

        if stop is not None:
            stop = Time(stop)

        if anchor is not None:
            anchor = Time(anchor)

        delay = float(delay)


        data['window'] = window

        if start is not None:
            data['start'] = start.subsec

        if stop is not None:
            data['stop'] = stop.subsec

        if anchor is not None:
            data['anchor'] = anchor.subsec

        data['delay'] = delay


        super().__init__(**data)



class WindowsParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    windows: Annotated[
        _WINDOWS,
        Field(...,
              description='Period for scheduling windows',
              min_length=0)]


    def __init__(
        self,
        windows: Optional[_WINDOWS] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        windows = windows or {}

        super().__init__(**{
            'windows': windows})



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path



def read_text(
    path: str | Path,
) -> str:
    """
    Read the text content from within the provided file path.

    .. testsetup::
       >>> tmpdir = getfixture('tmpdir')
       >>> path = Path(f'{tmpdir}/text.txt')

    Example
    -------
    >>> path.exists()
    False
    >>> save_text(path, 'foo')
    'foo'
    >>> read_text(path)
    'foo'

    :param path: Complete or relative path to the text file.
    :returns: Text content that was read from the file path.
    """

    path = Path(path).resolve()

    return path.read_text(
        encoding='utf-8')



def save_text(
    path: str | Path,
    content: str,
) -> str:
    """
    Save the provided text content to the provided file path.

    .. testsetup::
       >>> tmpdir = getfixture('tmpdir')
       >>> path = Path(f'{tmpdir}/text.txt')

    Example
    -------
    >>> path.exists()
    False
    >>> save_text(path, 'foo')
    'foo'
    >>> read_text(path)
    'foo'

    :param path: Complete or relative path to the text file.
    :param content: Content that will be written to the file.
    :returns: Text content that was read from the file path.
    """

    path = Path(path).resolve()

    path.write_text(
        data=content,
        encoding='utf-8')

    return read_text(path)



def append_text(
    path: str | Path,
    content: str,
) -> None:
    """
    Append the provided text content into provided file path.

    .. testsetup::
       >>> tmpdir = getfixture('tmpdir')
       >>> path = Path(f'{tmpdir}/text.txt')

    Example
    -------
    >>> path.exists()
    False
    >>> append_text(path, 'foo')
    >>> append_text(path, 'foo')
    >>> read_text(path)
    'foofoo'

    :param path: Complete or relative path to the text file.
    :param content: Content that will be written to the file.
    """

    path = Path(path).resolve()

    with path.open(
        mode='a',
        encoding='utf-8',
    ) as file:

        file.write(f'{content}')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dateutil.tz import gettz

from pytest import raises

from ..common import STAMP_SIMPLE
from ..common import UNIXEPOCH
from ..utils import findtz
from ..utils import strptime
from ..utils import utcdatetime



def test_utcdatetime() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    dtime = utcdatetime(1970, 1, 1)

    assert dtime.year == 1970
    assert dtime.month == 1
    assert dtime.day == 1
    assert dtime.hour == 0
    assert dtime.minute == 0
    assert dtime.second == 0


    dtime = utcdatetime(
        1970, 1, 1,
        tzinfo=gettz('US/Central'))

    assert dtime.year == 1970
    assert dtime.month == 1
    assert dtime.day == 1
    assert dtime.hour == 6
    assert dtime.minute == 0
    assert dtime.second == 0



def test_strptime() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    parsed = strptime(
        UNIXEPOCH, STAMP_SIMPLE)

    assert parsed.year == 1970
    assert parsed.month == 1
    assert parsed.day == 1
    assert parsed.hour == 0
    assert parsed.minute == 0
    assert parsed.second == 0



def test_strptime_raises() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    _raises = raises(ValueError)

    with _raises as reason:
        strptime('foo', '%Y')

    _reason = str(reason.value)

    assert _reason == 'invalid'



def test_findtz() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    tzinfo = findtz('UTC')

    assert 'UTC' in str(tzinfo)



def test_findtz_raises() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    _raises = raises(ValueError)

    with _raises as reason:
        findtz('foo')

    _reason = str(reason.value)

    assert _reason == 'tzname'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from fnmatch import fnmatch
from re import fullmatch as re_fmatch
from re import match as re_match



def rgxp_match(
    values: str | list[str],
    patterns: str | list[str],
    complete: bool = False,
) -> bool:
    """
    Determine whether or not values are included by patterns.

    Example
    -------
    >>> rgxp_match('one', ['on*', 'two'])
    True

    Example
    -------
    >>> rgxp_match('uno', ['on*', 'two'])
    False

    :param values: Value or values to enumerate for matching.
    :param patterns: Patterns which values can match any one.
    :param complete: Whether or not whole string must match.
    :returns: Boolean indicating outcome from the operation.
    """

    if isinstance(values, str):
        values = [values]

    if isinstance(patterns, str):
        patterns = [patterns]

    outcomes: set[bool] = set()

    function = (
        re_fmatch if complete
        else re_match)


    def _matches() -> bool:

        return bool(
            function(pattern, value))


    for value in values:

        matched = False

        for pattern in patterns:

            if not _matches():
                continue

            matched = True

            break

        outcomes.add(matched)


    return bool(
        outcomes and all(outcomes))



def fuzz_match(
    values: str | list[str],
    patterns: str | list[str],
) -> bool:
    """
    Determine whether or not values are included by patterns.

    Example
    -------
    >>> rgxp_match('one', ['on[a-z]', 'two'])
    True

    Example
    -------
    >>> rgxp_match('uno', ['on[a-z]', 'two'])
    False

    :param values: Value or values to enumerate for matching.
    :param patterns: Patterns which values can match any one.
    :returns: Boolean indicating outcome from the operation.
    """

    if isinstance(values, str):
        values = [values]

    if isinstance(patterns, str):
        patterns = [patterns]

    outcomes: set[bool] = set()


    for value in values:

        matched = False

        for pattern in patterns:

            if not fnmatch(value, pattern):
                continue

            matched = True

            break

        outcomes.add(matched)


    return bool(
        outcomes and all(outcomes))



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..common import STAMP_SIMPLE
from ..common import UNIXEPOCH
from ..common import UNIXHPOCH
from ..common import UNIXMPOCH
from ..time import Time
from ...types import inrepr
from ...types import instr
from ...types import lattrs



def test_Time() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    time = Time(
        UNIXEPOCH,
        format=STAMP_SIMPLE)


    attrs = lattrs(time)

    assert attrs == [
        '_Time__source']


    assert inrepr(
        "Time('1970-01-01T00:00",
        time)

    assert isinstance(
        hash(time), int)

    assert instr(
        '1970-01-01T00:00:00.000',
        time)


    assert int(time) == 0
    assert float(time) == 0.0

    assert time + 1 == Time(1)
    assert time - 1 == Time(-1)

    assert time == Time(0)
    assert time != Time(-1)
    assert time != 'invalid'

    assert time > Time(-1)
    assert time >= Time(0)
    assert time < Time(1)
    assert time <= Time(0)


    assert time.source.year == 1970

    assert time.epoch == 0.0

    assert time.spoch == 0

    assert time.mpoch == 0.0

    assert str(time.time) == '00:00:00'

    assert time.simple == UNIXEPOCH

    assert time.subsec == UNIXMPOCH

    assert time.human == UNIXHPOCH

    assert time.elapsed >= 1672531200

    assert time.since >= 1672531200

    assert time.before == (
        '1969-12-31T23:59:59.999999Z')

    assert time.after == (
        '1970-01-01T00:00:00.000001Z')

    assert time.stamp() == UNIXMPOCH

    time = time.shift('+1y')
    assert time == '1971-01-01'

    time = time.shifz('UTC-1')
    assert time == (
        '12/31/1970 23:00 -0100')



def test_Time_tzname() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    time1 = Time(
        '1970-01-01T00:00:00Z')

    time2 = time1.shifz(
        'US/Central')


    delta = time1 - time2

    assert -1 < delta < 1


    stamp1 = time1.stamp(
        tzname='US/Central')

    stamp2 = time2.stamp()

    assert stamp1 == stamp2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import copy
from datetime import datetime
from datetime import timedelta
from typing import Optional

from croniter import croniter

from .common import PARSABLE
from .common import SCHEDULE
from .parse import parse_time
from .time import Time



class Window:
    """
    Process and operate crontab or interval based schedule.

    .. testsetup::
       >>> from time import sleep

    Example
    -------
    >>> window = Window('* * * * *', '-4m@m')
    >>> [window.ready() for _ in range(6)]
    [True, True, True, True, True, False]

    :param window: Parameters for defining scheduled time.
    :param start: Determine the start for scheduling window.
    :param stop: Determine the ending for scheduling window.
    :param anchor: Optionally define time anchor for window.
    :param delay: Period of time schedulng will be delayed.
    """

    __window: SCHEDULE
    __start: Time
    __stop: Time
    __anchor: Time
    __delay: float

    __wlast: Time
    __wnext: Time


    def __init__(
        self,
        window: SCHEDULE,
        start: Optional[PARSABLE] = None,
        stop: Optional[PARSABLE] = None,
        *,
        anchor: Optional[PARSABLE] = None,
        delay: float = 0,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        anchor = anchor or start

        window = copy(window)
        start = Time(start)
        stop = Time(stop)
        anchor = Time(anchor)
        delay = float(delay)

        assert stop > start


        self.__window = window
        self.__start = start
        self.__stop = stop
        self.__anchor = anchor
        self.__delay = delay


        wlast, wnext = (
            self.__helper(anchor))

        while wnext > start:
            wlast, wnext = (
                self.__helper(wlast, True))

        while wnext < start:
            wlast, wnext = (
                self.__helper(wnext, False))

        self.__wlast = wlast
        self.__wnext = wnext



    @property
    def window(
        self,
    ) -> SCHEDULE:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return copy(self.__window)


    @property
    def start(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__start)


    @property
    def stop(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__stop)


    @property
    def anchor(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__anchor)


    @property
    def delay(
        self,
    ) -> float:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return float(self.__delay)


    @property
    def last(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__wlast)


    @property
    def next(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__wnext)


    @property
    def soonest(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """
        return Time(Time() - self.delay)


    @property
    def latest(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """
        return Time(self.stop - self.delay)


    @property
    def walked(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.next >= self.latest


    def __helper(
        self,
        anchor: PARSABLE,
        backward: bool = False,
    ) -> tuple[Time, Time]:
        """
        Determine next and last windows for window using anchor.

        :param anchor: Optionally define time anchor for window.
        :param backward: Optionally operate the window backward.
        :returns: Next and previous windows for schedule window.
        """

        window = self.__window

        if isinstance(window, str):
            return window_croniter(
                window, anchor, backward)

        if isinstance(window, dict):
            return window_interval(
                window, anchor, backward)

        raise NotImplementedError


    def ready(
        self,
        update: bool = True,
    ) -> bool:
        """
        Walk the internal time using current position in schedule.

        :param update: Determines whether or not time is updated.
        :returns: Boolean indicating outcome from the operation.
        """

        wnext = self.__wnext

        walked = self.walked
        soonest = self.soonest


        if walked is True:
            return False

        if wnext > soonest:
            return False


        wlast, wnext = (
            self.__helper(wnext))

        if update is True:
            self.__wlast = wlast
            self.__wnext = wnext


        return True


    def update(
        self,
        value: Optional[PARSABLE] = None,
    ) -> None:
        """
        Update the window from the provided parasable time value.

        :param value: Override the time updated for window value.
        """

        value = Time(
            value or 'now')

        wlast, wnext = (
            self.__helper(value))

        self.__wlast = wlast
        self.__wnext = wnext



def window_croniter(  # noqa: CFQ004
    schedule: str,
    anchor: PARSABLE,
    backward: bool = False,
) -> tuple[Time, Time]:
    """
    Determine next and previous times for cronjob schedule.

    :param window: Parameters for defining scheduled time.
    :param anchor: Optionally define time anchor for window.
    :param backward: Optionally operate the window backward.
    :returns: Next and previous windows for schedule window.
    """

    anchor = parse_time(anchor)


    def _wnext(
        source: datetime,
    ) -> datetime:

        parse = (
            _operator(source)
            .get_next())

        return parse_time(parse)


    def _wlast(
        source: datetime,
    ) -> datetime:

        parse = (
            _operator(source)
            .get_prev())

        return parse_time(parse)


    def _operator(
        source: datetime,
    ) -> croniter:
        return croniter(
            schedule, source)


    wnext = _wnext(anchor)

    if backward is True:
        wnext = _wlast(wnext)

    wlast = _wlast(wnext)


    return (
        Time(wlast),
        Time(wnext))



def window_interval(
    schedule: dict[str, int],
    anchor: PARSABLE,
    backward: bool = False,
) -> tuple[Time, Time]:
    """
    Determine next and previous times for interval schedule.

    :param window: Parameters for defining scheduled time.
    :param anchor: Optionally define time anchor for window.
    :param backward: Optionally operate the window backward.
    :returns: Next and previous windows for schedule window.
    """

    anpoch = (
        parse_time(anchor)
        .timestamp())

    seconds = (
        timedelta(**schedule)
        .seconds)


    if backward is True:
        wnext = anpoch
        wlast = anpoch - seconds

    else:
        wnext = anpoch + seconds
        wlast = anpoch


    return (
        Time(wlast),
        Time(wnext))



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..utils import config_load
from ..utils import config_path
from ..utils import config_paths
from ... import PROJECT
from ... import WORKSPACE



def test_config_load() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    loaded = config_load(
        'PROJECT/../.yamllint')

    assert list(loaded) == [
        'extends', 'rules']



def test_config_path() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    path = config_path('PROJECT')

    assert path == PROJECT


    path = config_path('WORKSPACE')

    assert path == WORKSPACE



def test_config_paths() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    paths = config_paths([
        'PROJECT', 'WORKSPACE'])

    assert paths == (
        PROJECT, WORKSPACE)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING

from .files import ConfigFiles
from .logger import Logger
from .params import Params
from .paths import ConfigPaths
from .utils import config_paths
from ..crypts import Crypts
from ..parse import Jinja2
from ..types import DictStrAny
from ..types import merge_dicts
from ..types import setate
from ..types import sort_dict

if TYPE_CHECKING:
    from ..utils.common import PATHABLE



class Config:
    """
    Contain the configurations from the arguments and files.

    .. note::
       Configuration loaded from files is validated with the
       Pydantic model :class:`encommon.config.Params`.

    .. testsetup::
       >>> from pathlib import Path
       >>> path = str(getfixture('tmpdir'))

    Example
    -------
    >>> config = Config()
    >>> config.config
    {'enconfig': None, 'encrypts': None, 'enlogger': None}

    :param files: Complete or relative path to config files.
    :param paths: Complete or relative path to config paths.
    :param cargs: Configuration arguments in dictionary form,
        which will override contents from the config files.
    :param sargs: Additional arguments on the command line.
    :param model: Override default config validation model.
    """

    __files: ConfigFiles
    __paths: ConfigPaths
    __cargs: DictStrAny
    __sargs: DictStrAny

    __model: Callable  # type: ignore

    __params: Optional[Params]
    __logger: Optional[Logger]
    __crypts: Optional[Crypts]
    __jinja2: Jinja2


    def __init__(
        self,
        files: Optional['PATHABLE'] = None,
        *,
        paths: Optional['PATHABLE'] = None,
        cargs: Optional[DictStrAny] = None,
        sargs: Optional[DictStrAny] = None,
        model: Optional[Callable] = None,  # type: ignore
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        files = files or []
        paths = paths or []
        cargs = cargs or {}
        sargs = sargs or {}

        paths = list(config_paths(paths))

        self.__model = model or Params
        self.__files = ConfigFiles(files)
        self.__cargs = deepcopy(cargs)
        self.__sargs = deepcopy(sargs)

        self.__params = None

        enconfig = (
            self.params.enconfig)

        if enconfig and enconfig.paths:
            paths.extend(enconfig.paths)

        self.__paths = ConfigPaths(paths)

        self.__logger = None
        self.__crypts = None

        jinja2 = Jinja2({
            'config': self})

        self.__jinja2 = jinja2


    @property
    def files(
        self,
    ) -> ConfigFiles:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__files


    @property
    def paths(
        self,
    ) -> ConfigPaths:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__paths


    @property
    def cargs(
        self,
    ) -> DictStrAny:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        returned: DictStrAny = {}

        cargs = deepcopy(self.__cargs)

        items = cargs.items()

        for key, value in items:
            setate(returned, key, value)

        return deepcopy(returned)


    @property
    def sargs(
        self,
    ) -> DictStrAny:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        returned: DictStrAny = {}

        sargs = deepcopy(self.__sargs)

        items = sargs.items()

        for key, value in items:
            setate(returned, key, value)

        return deepcopy(returned)


    @property
    def basic(
        self,
    ) -> DictStrAny:
        """
        Return the configuration source loaded from the objects.

        :returns: Configuration source loaded from the objects.
        """

        files = self.__files

        ferged = files.merge

        merge_dicts(
            dict1=ferged,
            dict2=self.cargs,
            force=True)

        return deepcopy(ferged)


    @property
    def merge(
        self,
    ) -> DictStrAny:
        """
        Return the configuration source loaded from the objects.

        :returns: Configuration source loaded from the objects.
        """

        files = self.__files
        paths = self.__paths

        ferged = files.merge
        perged = paths.merge

        merge_dicts(
            dict1=ferged,
            dict2=self.cargs,
            force=True)

        values = perged.values()

        for merge in values:

            merge_dicts(
                dict1=ferged,
                dict2=merge,
                force=False)

        return deepcopy(ferged)


    @property
    def model(
        self,
    ) -> Callable:  # type: ignore
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__model


    @property
    def params(
        self,
    ) -> Params:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        params = self.__params

        if params is not None:
            return params

        params = self.model(
            **self.basic)

        self.__params = params

        return self.__params


    @property
    def config(
        self,
    ) -> DictStrAny:
        """
        Return the configuration dumped from the Pydantic model.

        :returns: Configuration dumped from the Pydantic model.
        """

        return sort_dict(self.params.endumped)


    @property
    def logger(
        self,
    ) -> Logger:
        """
        Initialize the Python logging library using parameters.

        :returns: Instance of Python logging library created.
        """

        logger = self.__logger
        params = self.params

        if logger is not None:
            return logger

        logger = Logger(
            params.enlogger)

        self.__logger = logger

        return self.__logger


    @property
    def crypts(
        self,
    ) -> Crypts:
        """
        Initialize the encryption instance using the parameters.

        :returns: Instance of the encryption instance created.
        """

        crypts = self.__crypts
        params = self.params

        if crypts is not None:
            return crypts

        crypts = Crypts(
            params.encrypts)

        self.__crypts = crypts

        return self.__crypts


    @property
    def jinja2(
        self,
    ) -> Jinja2:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__jinja2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..hashes import Hashes
from ...types import inrepr
from ...types import instr
from ...types import lattrs



def test_Hashes() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    hashes = Hashes('string')


    attrs = lattrs(hashes)

    assert attrs == [
        '_Hashes__string']


    assert inrepr(
        'hashes.Hashes object',
        hashes)

    assert isinstance(
        hash(hashes), int)

    assert instr(
        'hashes.Hashes object',
        hashes)


    assert hashes.string == 'string'

    assert hashes.md5[:3] == 'b45'
    assert hashes.md5[-2:] == '21'

    assert hashes.sha1[:3] == 'ecb'
    assert hashes.sha1[-2:] == '4d'

    assert hashes.sha256[:3] == '473'
    assert hashes.sha256[-2:] == 'a8'

    assert hashes.sha512[:3] == '275'
    assert hashes.sha512[-2:] == '87'

    assert hashes.uuid[:3] == '38f'
    assert hashes.uuid[-2:] == 'e4'

    assert hashes.apache[:3] == '7LJ'
    assert hashes.apache[-2:] == '0='



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pydantic import BaseModel as Pydantic

from .types import DictStrAny



class BaseModel(Pydantic, extra='forbid'):
    """
    Pydantic base model but with added methods and routines.

    :param data: Keyword arguments passed to Pydantic model.
    """


    @property
    def endumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        return self.model_dump()


    @property
    def enpruned(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        return self.model_dump(
            exclude_none=True)



def clsname(
    cls: object,
) -> str:
    """
    Return the actual definition name for the Python class.

    :param cls: Provided Python class related to operation.
    :returns: Actual definition name for the Python class.
    """

    assert hasattr(
        cls, '__class__')

    _cls = cls.__class__

    assert hasattr(
        _cls, '__name__')

    return str(_cls.__name__)



def lattrs(
    cls: object,
) -> list[str]:
    """
    Return the list of attributes which are found in class.

    :param cls: Provided Python class related to operation.
    :returns: List of attributes which are found in class.
    """

    return list(cls.__dict__)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from time import sleep

from pytest import fixture
from pytest import raises

from ..params import TimerParams
from ..time import Time
from ..timers import Timers
from ...types import DictStrAny
from ...types import inrepr
from ...types import instr
from ...types import lattrs



@fixture
def timers(
    tmp_path: Path,
) -> Timers:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    model = TimerParams

    source: DictStrAny = {
        'one': model(timer=1),
        'two': model(timer=1)}


    store = (
        f'sqlite:///{tmp_path}'
        '/cache.db')

    timers = Timers(
        store=store)

    table = (
        timers
        .store_table)

    session = (
        timers
        .store_session)


    record = table(
        group='default',
        unique='two',
        last='1970-01-01T00:00:00Z',
        update='1970-01-01T00:00:00Z')

    session.merge(record)

    record = table(
        group='default',
        unique='tre',
        last='1970-01-01T00:00:00Z',
        update='1970-01-01T00:00:00Z')

    session.merge(record)

    session.commit()


    timers.create(
        'one', source['one'])

    timers.create(
        'two', source['two'])


    return timers



def test_Timers(
    timers: Timers,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param timers: Primary class instance for timers object.
    """


    attrs = lattrs(timers)

    assert attrs == [
        '_Timers__params',
        '_Timers__store',
        '_Timers__group',
        '_Timers__table',
        '_Timers__locker',
        '_Timers__sengine',
        '_Timers__session',
        '_Timers__childs']


    assert inrepr(
        'timers.Timers object',
        timers)

    assert isinstance(
        hash(timers), int)

    assert instr(
        'timers.Timers object',
        timers)


    assert timers.params

    assert timers.store[:6] == 'sqlite'

    assert timers.group == 'default'

    assert timers.store_table

    assert timers.store_engine

    assert timers.store_session

    assert len(timers.children) == 2


    timer = timers.children['one']

    assert timer.time >= Time('-1s')


    timer = timers.children['two']

    assert timer.time == '1970-01-01'



def test_Timers_cover(
    timers: Timers,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param timers: Primary class instance for timers object.
    """


    assert timers.ready('two')
    assert timers.pause('two')

    timers.update('two', 'now')

    assert not timers.ready('two')
    assert timers.pause('two')


    timers = Timers()


    params = TimerParams(timer=1)

    timers.create('fur', params)

    assert not timers.ready('fur')
    assert timers.pause('fur')

    sleep(1)

    assert timers.ready('fur')
    assert timers.pause('fur')

    timers.delete('fur')



def test_Timers_raises(
    timers: Timers,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param timers: Primary class instance for timers object.
    """


    _raises = raises(ValueError)

    with _raises as reason:
        timers.ready('dne')

    _reason = str(reason.value)

    assert _reason == 'unique'


    _raises = raises(ValueError)

    params = TimerParams(timer=1)

    with _raises as reason:
        timers.create('one', params)

    _reason = str(reason.value)

    assert _reason == 'unique'


    _raises = raises(ValueError)

    with _raises as reason:
        timers.update('dne', 'now')

    _reason = str(reason.value)

    assert _reason == 'unique'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .files import append_text
from .files import read_text
from .files import save_text
from .match import fuzz_match
from .match import rgxp_match
from .paths import resolve_path
from .paths import resolve_paths
from .paths import stats_path
from .sample import load_sample
from .sample import prep_sample
from .sample import read_sample
from .sample import rvrt_sample
from .stdout import array_ansi
from .stdout import kvpair_ansi
from .stdout import make_ansi
from .stdout import print_ansi
from .stdout import strip_ansi



__all__ = [
    'append_text',
    'array_ansi',
    'fuzz_match',
    'kvpair_ansi',
    'load_sample',
    'make_ansi',
    'prep_sample',
    'print_ansi',
    'read_sample',
    'read_text',
    'rvrt_sample',
    'resolve_path',
    'resolve_paths',
    'rgxp_match',
    'save_text',
    'stats_path',
    'strip_ansi']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional

from .common import NUMERIC
from .common import PARSABLE
from .time import Time



class Timer:
    """
    Process and operate the timer using provided interval.

    .. testsetup::
       >>> from time import sleep

    Example
    -------
    >>> timer = Timer(1)
    >>> timer.ready()
    False
    >>> sleep(1)
    >>> timer.ready()
    True

    :param seconds: Period of time which must pass in timer.
    :param started: Override default of using current time.
    """

    __timer: float
    __time: Time


    def __init__(
        self,
        timer: NUMERIC,
        *,
        start: Optional[PARSABLE] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        timer = float(timer)
        start = Time(start)

        self.__timer = timer
        self.__time = start


    @property
    def timer(
        self,
    ) -> float:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__timer


    @property
    def time(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__time


    @property
    def since(
        self,
    ) -> float:
        """
        Return the seconds that have elapsed since the interval.

        :returns: Seconds that have elapsed since the interval.
        """

        return self.time.since


    @property
    def remains(
        self,
    ) -> float:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        timer = self.timer
        since = self.since

        return max(0.0, timer - since)


    def ready(
        self,
        update: bool = True,
    ) -> bool:
        """
        Determine whether or not the appropriate time has passed.

        :param update: Determines whether or not time is updated.
        :returns: Boolean indicating whether enough time passed.
        """

        if self.remains > 0:
            return False

        if update is True:
            self.update('now')

        return True


    def pause(
        self,
        update: bool = True,
    ) -> bool:
        """
        Determine whether or not the appropriate time has passed.

        :param update: Determines whether or not time is updated.
        :returns: Boolean indicating whether enough time passed.
        """

        return not self.ready(update)


    def update(
        self,
        value: Optional[PARSABLE] = None,
    ) -> None:
        """
        Update the timer from the provided parasable time value.

        :param value: Override the time updated for timer value.
        """

        value = Time(
            value or 'now')

        self.__time = value



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pytest import raises

from ..network import Network
from ..network import insubnet_ip
from ..network import isvalid_ip
from ...types import inrepr
from ...types import instr
from ...types import lattrs



def test_Network() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    naddr = Network('12.34.56.7')


    attrs = lattrs(naddr)

    assert attrs == [
        '_Network__source']


    assert inrepr(
        "Network('12.34.",
        naddr)

    assert isinstance(
        hash(naddr), int)

    assert instr(
        '12.34.56.7/32',
        naddr)


    assert naddr == '12.34.56.7'
    assert naddr != 'invalid'


    assert naddr.source

    assert naddr.version == 4

    assert naddr.cidr == 32

    assert naddr.address == '12.34.56.7'

    assert naddr.address_cidr == '12.34.56.7/32'

    assert naddr.address_host == '12.34.56.7/32'

    assert naddr.network == '12.34.56.7'

    assert naddr.network_cidr == '12.34.56.7/32'

    assert not naddr.broadcast

    assert naddr.netmask == '255.255.255.255'

    assert naddr.padded == '012.034.056.007'

    assert naddr.reverse == '7.56.34.12'

    assert naddr.hwaddr == '01-20-34-05-60-07'

    assert naddr.ispublic

    assert not naddr.isprivate

    assert not naddr.islinklocal

    assert not naddr.islocalhost



def test_Network_ipv4() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    naddr = Network('12.34.56.7/24')


    assert naddr.cidr == 24

    assert naddr.address == '12.34.56.7'

    assert naddr.address_cidr == '12.34.56.7/24'

    assert naddr.address_host == '12.34.56.7/32'

    assert naddr.network == '12.34.56.0'

    assert naddr.network_cidr == '12.34.56.0/24'

    assert naddr.broadcast == '12.34.56.255'

    assert naddr.netmask == '255.255.255.0'

    assert naddr.padded == '012.034.056.007'

    assert naddr.reverse == '7.56.34.12'

    assert naddr.hwaddr == '01-20-34-05-60-07'

    assert naddr.ispublic

    assert not naddr.isprivate

    assert not naddr.islinklocal

    assert not naddr.islocalhost



def test_Network_ipv6() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    naddr = Network('2001:db8::/64')


    assert naddr.cidr == 64

    assert naddr.address == '2001:db8::'

    assert naddr.address_cidr == '2001:db8::/64'

    assert naddr.address_host == '2001:db8::/128'

    assert naddr.network == '2001:db8::'

    assert naddr.network_cidr == '2001:db8::/64'

    assert naddr.broadcast == (
        '2001:db8::ffff:ffff:ffff:ffff')

    assert naddr.netmask == (
        'ffff:ffff:ffff:ffff::')

    with raises(ValueError):
        naddr.padded

    with raises(ValueError):
        naddr.reverse

    with raises(ValueError):
        naddr.hwaddr

    assert not naddr.ispublic

    assert naddr.isprivate

    assert not naddr.islinklocal

    assert not naddr.islocalhost



def test_isvalid_ip() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert isvalid_ip('1.2.3.4')

    assert not isvalid_ip('1.2')

    assert not isvalid_ip(None)



def test_insubnet_ip() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert insubnet_ip(
        '192.168.1.1',
        ['172.16.0.0/12',
         '192.168.0.0/16'])

    assert not insubnet_ip(
        '1.2.3.4',
        ['172.16.0.0/12',
         '192.168.0.0/16'])

    assert insubnet_ip(
        '192.168.1.1',
        ['192.168.1.1/32'])

    assert insubnet_ip(
        '192.168.1.1/32',
        '192.168.1.1/32')

    assert insubnet_ip(
        '2001:db8::',
        ['2001:db8::/128'])

    assert insubnet_ip(
        '2001:db8::',
        ['2001:db8::/32'])

    assert insubnet_ip(
        '2001:db8::/128',
        ['2001:db8::/32'])

    with raises(ValueError):

        insubnet_ip(
            'invalid',
            ['2001:db8::/32'])



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from selenium.webdriver.remote.webdriver import WebDriver

from . import _executejs

if TYPE_CHECKING:
    from ..content import Content



def test_datagrid(
    content: 'Content',
    driver: WebDriver,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param content: Primary class instance for the content.
    :param driver: Selenium driver provided by the library.
    """


    driver.get(
        f"""
        data:text/html;
        charset=utf-8,
        <html>
          <head>
            <script>
        {content.scripts('default')}
        {content.scripts('helpers')}
        {content.scripts('datagrid')}
            </script>
          </head>
          <body></body>
        </html>
        """)


    fields = {
        'foo': 'Foo',
        'bar': 'Bar'}

    values = [
        {'foo': '1',
         'bar': 1},
        {'foo': '2',
         'bar': 2}]


    outcome = _executejs(
        driver,
        'let element ='
        ' datagrid(...arguments);'
        'return'
        ' element[0].outerHTML;',
        fields,
        values,
        wrap=False)


    assert outcome == (
        '<table class="encommon_datagrid">'
        '<thead>'
        '<tr>'
        '<th>Bar</th>'
        '<th>Foo</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr>'
        '<td>1</td>'
        '<td>1</td>'
        '</tr>'
        '<tr>'
        '<td>2</td>'
        '<td>2</td>'
        '</tr>'
        '</tbody>'
        '</table>')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from json import dumps
from json import loads
from pathlib import Path

from ..sample import load_sample
from ..sample import prep_sample
from ..sample import read_sample
from ..sample import rvrt_sample
from ..stdout import ArrayColors
from ... import PROJECT
from ...config import LoggerParams
from ...utils.sample import ENPYRWS



_PREFIX = 'encommon_sample'

_SOURCE = {
    'list': ['bar', 'baz'],
    'tuple': (1, 2),
    'project': PROJECT,
    'other': '/pat/h',
    'devnull': '/dev/null'}

_EXPECT = {
    'list': ['bar', 'baz'],
    'tuple': [1, 2],
    'project': f'_/{_PREFIX}/PROJECT/_',
    'other': f'_/{_PREFIX}/pytemp/_',
    'devnull': '/dev/null'}

_REPLACES = {
    'PROJECT': str(PROJECT),
    'pytemp': '/pat/h'}



def test_prep_sample() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    colors = ArrayColors()

    params = LoggerParams()

    source = {
        'colors': colors,
        'params': params}

    sample = prep_sample(
        content=source,
        indent=None)

    assert loads(sample) == {
        'colors': {
            'bool': 93,
            'colon': 37,
            'empty': 36,
            'hyphen': 37,
            'key': 97,
            'label': 37,
            'none': 33,
            'num': 93,
            'other': 91,
            'str': 92,
            'times': 96},
        'params': {
            'file_level': None,
            'file_path': None,
            'stdo_level': None}}



def test_load_sample(
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    """

    source = deepcopy(_SOURCE)

    expect = deepcopy(_EXPECT)


    sample_path = (
        tmp_path / 'samples.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=source,
        replace=_REPLACES)

    _expect = dumps(
        expect, indent=2)

    assert _expect == sample


    source |= {'list': [1, 3, 2]}
    expect |= {'list': [1, 3, 2]}

    sample = load_sample(
        path=sample_path,
        content=source,
        update=True,
        replace=_REPLACES)

    _expect = dumps(
        expect, indent=2)

    assert _expect == sample



def test_read_sample(
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    """

    source = prep_sample(
        content=_SOURCE,
        indent=None)

    sample = read_sample(
        sample=source,
        replace=_REPLACES)

    sample = prep_sample(
        content=loads(sample),
        replace=_REPLACES,
        indent=None)

    expect = dumps(_EXPECT)

    assert expect == sample



def test_rvrt_sample(
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    """

    source = prep_sample(
        content=_SOURCE,
        indent=None)

    sample = rvrt_sample(
        sample=source,
        replace=_REPLACES)

    sample = prep_sample(
        content=loads(sample),
        replace=_REPLACES,
        indent=None)

    expect = dumps(_EXPECT)

    assert expect == sample



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Sequence



def dedup_list(
    value: list[Any],
    update: bool = True,
) -> list[Any]:
    """
    Return the provided list values with duplicates removed.

    .. warning::
       This function will update the ``value`` reference.

    Example
    -------
    >>> value = [1, 2, 2, 3]
    >>> dedup_list(value)
    [1, 2, 3]
    >>> value
    [1, 2, 3]

    :param value: List of values processed for duplication.
    :param update: Indicate if to update provided reference.
    :returns: Provided list values with duplicates removed.
    """

    if update is False:
        value = list(value)


    unique: list[Any] = []

    for item in list(value):

        if item in unique:

            value.remove(item)

            continue

        unique.append(item)


    return value



def fuzzy_list(
    values: str | list[str],
    patterns: str | list[str],
) -> list[str]:
    """
    Return the provided values that match provided patterns.

    Example
    -------
    >>> values = ['foo', 'bar', 'baz', 'bop']
    >>> patterns = ['*o', 'ba*']
    >>> fuzzy_list(values, patterns)
    ['foo', 'bar', 'baz']

    :param values: Value or values to enumerate for matching.
    :param patterns: Patterns which values can match any one.
    :returns: Provided values that match provided patterns.
    """

    from ..utils import fuzz_match

    if isinstance(values, str):
        values = [values]

    matched: list[str] = []

    for value in values:

        match = fuzz_match(
            value, patterns)

        if match is False:
            continue

        matched.append(value)

    return matched



def inlist(
    needle: Any,  # noqa: ANN401
    haystack: Sequence[Any],
) -> bool:
    """
    Return the boolean indicating whether needle in haystack.

    Example
    -------
    >>> haystack = [1, 2, 3]
    >>> inlist(2, haystack)
    True

    :param needle: Provided item that may be within haystack.
    :param haystack: List of items which may contain needle.
    :returns: Boolean indicating whether needle in haystack.
    """

    return needle in haystack



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from pytest import fixture

from selenium.webdriver.remote.webdriver import WebDriver

from . import _executejs

if TYPE_CHECKING:
    from ..content import Content



@fixture
def prepare(
    content: 'Content',
    driver: WebDriver,
) -> tuple['Content', WebDriver]:
    """
    Construct the instance for use in the downstream tests.

    :returns: Newly constructed instance of related class.
    """

    driver.get(
        f"""
        data:text/html;
        charset=utf-8,
        <html>
          <head>
            <script>
        {content.scripts('default')}
        {content.scripts('helpers')}
        {content.scripts('numeric')}
            </script>
          </head>
          <body>
            <div>Hello</div>
            <span>World</span>
          </body>
        </html>
        """)

    return (content, driver)



def test_numeric(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare


    outcome = _executejs(
        driver,
        'let element ='
        ' numeric(...arguments);'
        'return'
        ' element[0].outerHTML;',
        1.1, 'gb',
        wrap=False)

    assert outcome == (
        '<div class="encommon_numeric">'
        '<span class="_value">1</span>'
        '<span class="_delim">.</span>'
        '<span class="_decimal">1</span>'
        '<span class="_unit">gb</span>'
        '</div>')


    outcome = _executejs(
        driver,
        'let element ='
        ' numeric(...arguments);'
        'return'
        ' element[0].outerHTML;',
        1.0, 'gb',
        wrap=False)

    assert outcome == (
        '<div class="encommon_numeric">'
        '<span class="_value">1</span>'
        '<span class="_unit">gb</span>'
        '</div>')



def test_numeric_count(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare


    outcome = _executejs(
        driver,
        'let element ='
        ' numeric_count(...arguments);'
        'return'
        ' element[0].outerHTML;',
        1e10 + 1e8,
        wrap=False)


    assert outcome == (
        '<div class="encommon_numeric">'
        '<span class="_value">10</span>'
        '<span class="_delim">.</span>'
        '<span class="_decimal">1</span>'
        '<span class="_unit">billion</span>'
        '</div>')



def test_numeric_bytes(
    prepare: tuple['Content', WebDriver],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param prepare: Driver and content loaded with scripts.
    """

    content, driver = prepare


    outcome = _executejs(
        driver,
        'let element ='
        ' numeric_bytes(...arguments);'
        'return'
        ' element[0].outerHTML;',
        1e10 + 1e8,
        wrap=False)


    assert outcome == (
        '<div class="encommon_numeric">'
        '<span class="_value">10</span>'
        '<span class="_delim">.</span>'
        '<span class="_decimal">1</span>'
        '<span class="_unit">GB</span>'
        '</div>')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

from .utils import config_load
from .utils import config_path
from .utils import config_paths
from ..types import DictStrAny
from ..types import merge_dicts
from ..types import sort_dict

if TYPE_CHECKING:
    from ..utils.common import PATHABLE



class ConfigFile:
    """
    Contain the configuration content from filesystem path.

    :param path: Complete or relative path to configuration.
    """

    path: Path
    config: DictStrAny


    def __init__(
        self,
        path: str | Path,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.path = config_path(path)
        self.config = config_load(path)



class ConfigFiles:
    """
    Enumerate files and store the contents on relative path.

    .. note::
       Class can be empty in order to play nice with parent.

    :param paths: Complete or relative path to config files.
    :param force: Force the merge on earlier files by later.
    """

    paths: tuple[Path, ...]
    config: dict[str, ConfigFile]

    __merge: Optional[DictStrAny]


    def __init__(
        self,
        paths: 'PATHABLE',
        force: bool = False,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.paths = config_paths(paths)

        self.config = {
            str(x): ConfigFile(x)
            for x in self.paths}

        self.__merge = None


    @property
    def merge(
        self,
    ) -> DictStrAny:
        """
        Return the configuration in dictionary format for files.

        :returns: Configuration in dictionary format for files.
        """

        config = self.config
        merge = self.__merge

        if merge is not None:
            return deepcopy(merge)

        merge = {}


        for file in config.values():

            source = file.config

            merge_dicts(
                dict1=merge,
                dict2=deepcopy(source),
                force=False)


        merge = sort_dict(merge)

        self.__merge = merge

        return deepcopy(merge)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path



PROJECT = Path(__file__).parent
WORKSPACE = PROJECT.parents[2]

VERSION = (
    (PROJECT / 'version.txt')
    .read_text(encoding='utf-8')
    .splitlines()[0].strip())



__version__ = VERSION



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

from yaml import SafeLoader
from yaml import load

from .. import PROJECT
from .. import WORKSPACE
from ..types import DictStrAny
from ..utils import read_text
from ..utils import resolve_path
from ..utils import resolve_paths

if TYPE_CHECKING:
    from ..utils.common import PATHABLE
    from ..utils.common import REPLACE



def config_load(
    path: str | Path,
) -> DictStrAny:
    """
    Load configuration using the directory or file provided.

    :param path: Complete or relative path to configuration.
    :returns: New resolved filesystem path object instance.
    """

    loaded = read_text(
        config_path(path))

    parsed = load(loaded, SafeLoader)

    assert isinstance(parsed, dict)

    return parsed



def config_path(
    path: str | Path,
) -> Path:
    """
    Resolve the provided path replacing the magic keywords.

    .. note::
       This function simply wraps one from utils subpackage.

    :param path: Complete or relative path for processing.
    :returns: New resolved filesystem path object instance.
    """

    replace = {
        'PROJECT': PROJECT,
        'WORKSPACE': WORKSPACE}

    return resolve_path(path, replace)



def config_paths(
    paths: 'PATHABLE',
    replace: Optional['REPLACE'] = None,
) -> tuple[Path, ...]:
    """
    Resolve the provided paths replacing the magic keywords.

    .. note::
       This function simply wraps one from utils subpackage.

    :param paths: Complete or relative paths for processing.
    :param replace: Optional values to replace in the path.
    :returns: New resolved filesystem path object instances.
    """

    if replace is None:
        replace = {}

    replace = dict(replace)

    replace = {
        'PROJECT': PROJECT,
        'WORKSPACE': WORKSPACE}

    return resolve_paths(
        paths, replace)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any



COMMAS = ', '
COMMAD = ','

NEWLINE = '\n'

SEMPTY = ''
SPACED = ' '



def strplwr(
    value: Any,  # noqa: ANN401
) -> str:
    """
    Return the provided string but stripped and lower cased.

    Example
    -------
    >>> strplwr('  Foo ')
    'foo'

    :param value: String which will be stripped and lowered.
    :returns: Provided string but stripped and lower cased.
    """

    _value = str(value)

    return _value.strip().lower()



def hasstr(
    haystack: str,
    needle: str,
) -> bool:
    """
    Return the boolean indicating if needle is in haystack.

    Example
    -------
    >>> haystack = 'barfoobaz'
    >>> needle = 'foo'
    >>> hasstr(haystack, needle)
    True

    :param haystack: Full string which may contain needle.
    :param needle: Substring which may be within haystack.
    :returns: Boolean indicating if needle is in haystack.
    """

    return needle in haystack



def inrepr(
    needle: str,
    haystack: Any,  # noqa: ANN401
) -> bool:
    """
    Return the boolean indicating if needle is in haystack.

    Example
    -------
    >>> class MyClass: pass
    >>> haystack = MyClass()
    >>> needle = 'MyClass'
    >>> inrepr(needle, haystack)
    True

    :param needle: Substring which may be within haystack.
    :param haystack: Object which is inspected for needle.
    :returns: Boolean indicating if needle is in haystack.
    """

    return hasstr(repr(haystack), needle)



def instr(
    needle: str,
    haystack: Any,  # noqa: ANN401
) -> bool:
    """
    Return the boolean indicating if needle is in haystack.

    Example
    -------
    >>> class MyClass: pass
    >>> haystack = MyClass()
    >>> needle = 'MyClass'
    >>> instr(needle, haystack)
    True

    :param needle: Substring which may be within haystack.
    :param haystack: Object which is inspected for needle.
    :returns: Boolean indicating if needle is in haystack.
    """

    return hasstr(str(haystack), needle)



def rplstr(
    source: str,
    match: str,
    value: Any,  # noqa: ANN401
) -> str:
    """
    Return the source string with the match value replaced.

    Example
    -------
    >>> rplstr('foo', 'foo', 'bar')
    'bar'

    :param source: String that to be processed and returned.
    :param match: What will be replaced within the string.
    :param value: Replace value for the string is matched.
    :returns: Source string with the match value replaced.
    """

    match = str(match)
    value = str(value)

    return (
        str(source)
        .replace(match, value))



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from . import SAMPLES
from ... import PROJECT
from ...types import inrepr
from ...types import instr
from ...types import lattrs
from ...utils import load_sample
from ...utils import prep_sample
from ...utils.sample import ENPYRWS

if TYPE_CHECKING:
    from ..config import Config



def test_Config(  # noqa: CFQ001
    tmp_path: Path,
    config: 'Config',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param config: Primary class instance for configuration.
    """


    attrs = lattrs(config)

    assert attrs == [
        '_Config__model',
        '_Config__files',
        '_Config__cargs',
        '_Config__sargs',
        '_Config__params',
        '_Config__paths',
        '_Config__logger',
        '_Config__crypts',
        '_Config__jinja2']


    assert inrepr(
        'config.Config object',
        config)

    assert isinstance(
        hash(config), int)

    assert instr(
        'config.Config object',
        config)


    assert config.files.merge

    assert config.paths.merge

    assert len(config.cargs) == 1

    assert len(config.sargs) == 1

    assert len(config.basic) == 3

    assert len(config.merge) == 5

    assert callable(config.model)

    assert config.params

    assert len(config.config) == 3

    assert config.logger

    assert config.crypts

    assert config.jinja2


    replaces = {
        'pytemp': tmp_path,
        'PROJECT': PROJECT}


    sample_path = (
        SAMPLES / 'basic.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=config.basic,
        replace=replaces)

    expect = prep_sample(
        content=config.basic,
        replace=replaces)

    assert expect == sample


    sample_path = (
        SAMPLES / 'merge.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=config.merge,
        replace=replaces)

    expect = prep_sample(
        content=config.merge,
        replace=replaces)

    assert expect == sample


    sample_path = (
        SAMPLES / 'config.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=config.config,
        replace=replaces)

    expect = prep_sample(
        content=config.config,
        replace=replaces)

    assert expect == sample



def test_Config_jinja2(
    config: 'Config',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param config: Primary class instance for configuration.
    """

    jinja2 = config.jinja2
    j2parse = jinja2.parse

    parsed = j2parse(
        '{{ foo }}',
        {'foo': 'bar'})

    assert parsed == 'bar'



def test_Config_cover(
    config: 'Config',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param config: Primary class instance for configuration.
    """


    logger1 = config.logger
    logger2 = config.logger

    assert logger1 is logger2


    crypts1 = config.crypts
    crypts2 = config.crypts

    assert crypts1 is crypts2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .classes import BaseModel
from .classes import clsname
from .classes import lattrs
from .dicts import merge_dicts
from .dicts import sort_dict
from .empty import Empty
from .funcs import funcname
from .lists import dedup_list
from .lists import fuzzy_list
from .lists import inlist
from .notate import delate
from .notate import expate
from .notate import getate
from .notate import impate
from .notate import setate
from .strings import hasstr
from .strings import inrepr
from .strings import instr
from .strings import rplstr
from .strings import strplwr
from .types import DictStrAny
from .types import LDictStrAny
from .types import NCFalse
from .types import NCNone
from .types import NCTrue



__all__ = [
    'BaseModel',
    'clsname',
    'dedup_list',
    'fuzzy_list',
    'delate',
    'DictStrAny',
    'Empty',
    'expate',
    'funcname',
    'getate',
    'hasstr',
    'impate',
    'inlist',
    'inrepr',
    'instr',
    'lattrs',
    'LDictStrAny',
    'merge_dicts',
    'rplstr',
    'setate',
    'sort_dict',
    'strplwr',
    'NCTrue',
    'NCFalse',
    'NCNone']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any

from selenium.webdriver.remote.webdriver import WebDriver



def _executejs(
    driver: WebDriver,
    script: str,
    *args: Any,
    wrap: bool = True,
) -> Any:  # noqa: ANN401
    """
    Evaluate the provided JavaScript using the test driver.

    :param driver: Selenium driver provided by the library.
    :param script: Actual code that is executed with engine.
    :param args: Positional arguments passed for downstream.
    :param wrap: Determines if script is wrapped in return.
    """

    script = (
        f'return {script};'
        if wrap else script)

    result = (
        driver
        .execute_script(script, *args))

    return result



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from ..types import BaseModel



_CRYPTS = dict[str, 'CryptParams']



class CryptParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    phrase: Annotated[
        str,
        Field(...,
              description='Passphrase for the operations',
              min_length=1)]


    def __init__(
        self,
        phrase: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**{
            'phrase': phrase})



class CryptsParams(BaseModel, extra='forbid'):
    """
    Process and validate the core configuration parameters.
    """

    phrases: Annotated[
        _CRYPTS,
        Field(...,
              description='Passphrases for the operations',
              min_length=0)]


    def __init__(
        self,
        phrases: Optional[_CRYPTS] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        phrases = phrases or {}

        super().__init__(**{
            'phrases': phrases})



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from base64 import b64encode
from hashlib import md5
from hashlib import sha1
from hashlib import sha256
from hashlib import sha512
from uuid import NAMESPACE_DNS
from uuid import uuid3



class Hashes:
    """
    Create hash values using the provided at instantiation.

    Example
    -------
    >>> hashes = Hashes('string')
    >>> hashes.sha256
    '473287f8298dba7163a897908958f7c0ea...

    Example
    -------
    >>> hashes = Hashes('string')
    >>> hashes.uuid
    '38ffd1ed-2b4d-3c84-ae45-bf3bf354eee4'

    :param string: String which will be used within hashing.
    """

    __string: str


    def __init__(
        self,
        string: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__string = string


    @property
    def string(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__string


    @property
    def md5(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        encode = self.__string.encode()

        return md5(encode).hexdigest()


    @property
    def sha1(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        string = self.__string.encode()

        return sha1(string).hexdigest()


    @property
    def sha256(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        string = self.__string.encode()

        return sha256(string).hexdigest()


    @property
    def sha512(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        string = self.__string.encode()

        return sha512(string).hexdigest()


    @property
    def uuid(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        string = self.__string

        return str(
            uuid3(NAMESPACE_DNS, string))


    @property
    def apache(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        string = self.__string.encode()

        digest = sha1(string).digest()

        return b64encode(digest).decode()



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from json import dumps
from logging import CRITICAL
from logging import DEBUG
from logging import ERROR
from logging import FileHandler
from logging import Formatter
from logging import INFO
from logging import Logger as _Logger
from logging import NOTSET
from logging import NullHandler
from logging import StreamHandler
from logging import WARNING
from logging import getLogger
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from .utils import config_path
from ..times import Time
from ..types import Empty
from ..types.strings import COMMAD
from ..types.strings import COMMAS
from ..types.strings import SPACED
from ..utils import kvpair_ansi
from ..utils.common import JOINABLE

if TYPE_CHECKING:
    from .params import LoggerParams
    from ..times.common import PARSABLE



LOGR_FILE = 'encommon.logger.file'
LOGR_STDO = 'encommon.logger.stdo'

LOGLEVELS = Literal[
    'critical',
    'debug',
    'error',
    'info',
    'warning']

LOGSEVERS = {
    'critical': int(CRITICAL),
    'debug': int(DEBUG),
    'error': int(ERROR),
    'info': int(INFO),
    'warning': int(WARNING)}



class Message:
    """
    Format the provided keyword arguments for logging output.

    .. note::
       Log messages are expected to contain string or numeric.

    .. testsetup::
       >>> from encommon.utils.stdout import strip_ansi

    Example
    -------
    >>> message = Message('info', '1970-01-01', foo='bar')
    >>> strip_ansi(message.stdo_output)
    'level="info" time="1970-01-01T00:00:00Z" foo="bar"'

    :param level: Severity which log message is classified.
    :param time: What time the log message actually occurred.
    :param kwargs: Keyword arguments for populating message.
    """

    __level: LOGLEVELS
    __time: Time
    __fields: dict[str, str] = {}


    def __init__(
        self,
        level: LOGLEVELS,
        time: Optional['PARSABLE'] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__level = level
        self.__time = Time(time)
        self.__fields = {}

        items = kwargs.items()

        for key, value in items:


            if (hasattr(value, '__len__')
                    and not len(value)):
                continue

            if value in [None, Empty]:
                continue


            if isinstance(value, JOINABLE):

                values = sorted(
                    str(x) for x in value)

                value = COMMAD.join(values)


            if (isinstance(value, Time)
                    and key == 'elapsed'):
                value = value.since


            if (isinstance(value, float)
                    and key == 'elapsed'):

                value = round(value, 2)


            value = str(value).strip()

            self.__fields[key] = value


    def __repr__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        fields: dict[str, str] = {
            'level': self.__level,
            'time': self.__time.subsec}

        fields |= dict(self.__fields)

        message = COMMAS.join([
            f'{k}="{v}"' for k, v
            in fields.items()])

        return f'Message({message})'


    def __str__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return self.__repr__()


    @property
    def level(
        self,
    ) -> LOGLEVELS:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__level


    @property
    def time(
        self,
    ) -> Time:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return Time(self.__time)


    @property
    def fields(
        self,
    ) -> dict[str, str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__fields)


    @property
    def stdo_output(
        self,
    ) -> str:
        """
        Format keyword arguments for writing to standard output.

        :returns: String representation for the standard output.
        """

        fields: dict[str, str] = {
            'level': self.__level,
            'time': self.__time.simple}

        fields |= dict(self.__fields)


        fields['time'] = (
            fields['time']
            .replace('+0000', 'Z'))


        output: list[str] = []

        items = fields.items()

        for field, value in items:

            _value = kvpair_ansi(
                field, value)

            output.append(_value)

        return SPACED.join(output)


    @property
    def file_output(
        self,
    ) -> str:
        """
        Format keyword arguments for writing to filesystem path.

        :returns: String representation for the filesystem path.
        """

        fields: dict[str, str] = {
            'level': self.__level,
            'time': self.__time.subsec}

        fields |= dict(self.__fields)


        return dumps(fields)



class FileFormatter(Formatter):
    """
    Supplement class for built-in logger exception formatter.

    .. note::
       Input parameters are not defined, check parent class.
    """


    def formatException(
        self,
        ei: Any,  # noqa: ANN401
    ) -> str:
        """
        Specifically overrides method for formatting exceptions.

        :param ei: Exception information provided by the logger.
        :returns: String representation for the filesystem path.
        """

        reason = super().formatException(ei)

        message = Message(
            level='error',
            status='exception',
            reason=reason)

        return message.file_output



class Logger:
    """
    Manage the file and standard output with logging library.

    .. note::
       Uses keyword name for levels in Pyton logging library.

    +-----------+-----------+
    | *Numeric* | *Keyword* |
    +-----------+-----------+
    | 10        | debug     |
    +-----------+-----------+
    | 20        | info      |
    +-----------+-----------+
    | 30        | warning   |
    +-----------+-----------+
    | 40        | error     |
    +-----------+-----------+
    | 50        | critical  |
    +-----------+-----------+

    Example
    -------
    >>> logger = Logger(stdo_level='info')
    >>> logger.start()
    >>> logger.log_i(message='testing')

    :param stdo_level: Minimum level for the message to pass.
    :param file_level: Minimum level for the message to pass.
    :param file_path: Enables writing to the filesystem path.
    :param params: Parameters used to instantiate the class.
    """

    __params: 'LoggerParams'

    __stdo_level: Optional[LOGLEVELS]
    __file_level: Optional[LOGLEVELS]
    __file_path: Optional[Path]

    __started: bool

    __logr_stdo: _Logger
    __logr_file: _Logger


    def __init__(
        self,
        params: Optional['LoggerParams'] = None,
        *,
        stdo_level: Optional[LOGLEVELS] = None,
        file_level: Optional[LOGLEVELS] = None,
        file_path: Optional[str | Path] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        from .params import LoggerParams

        if params is None:
            params = LoggerParams(
                stdo_level=stdo_level,
                file_level=file_level,
                file_path=file_path)

        self.__params = deepcopy(params)

        stdo_level = params.stdo_level
        file_level = params.file_level
        file_path = params.file_path

        if file_path is not None:
            file_path = config_path(file_path)

        self.__stdo_level = stdo_level
        self.__file_level = file_level
        self.__file_path = file_path

        self.__started = False

        logr_stdo = getLogger(LOGR_STDO)
        logr_file = getLogger(LOGR_FILE)

        self.__logr_stdo = logr_stdo
        self.__logr_file = logr_file


    @property
    def params(
        self,
    ) -> 'LoggerParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def stdo_level(
        self,
    ) -> Optional[LOGLEVELS]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__stdo_level


    @property
    def file_level(
        self,
    ) -> Optional[LOGLEVELS]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__file_level


    @property
    def file_path(
        self,
    ) -> Optional[Path]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__file_path


    @property
    def started(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__started


    @property
    def logger_stdo(
        self,
    ) -> _Logger:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__logr_stdo


    @property
    def logger_file(
        self,
    ) -> _Logger:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__logr_file


    def start(
        self,
    ) -> None:
        """
        Initialize the Python logging library using parameters.
        """

        stdo_level = self.__stdo_level
        file_level = self.__file_level
        file_path = self.__file_path

        logr_stdo = self.__logr_stdo
        logr_file = self.__logr_file


        logr_root = getLogger()

        logr_root.setLevel(NOTSET)

        logr_stdo.handlers = [NullHandler()]
        logr_file.handlers = [NullHandler()]


        if stdo_level is not None:

            level = LOGSEVERS[stdo_level]

            handstdo = StreamHandler()

            format = Formatter(
                '%(message)s')

            logr_stdo.handlers = [handstdo]

            handstdo.setLevel(level)
            handstdo.setFormatter(format)


        if file_path and file_level:

            level = LOGSEVERS[file_level]

            handfile = (
                FileHandler(file_path))

            format = FileFormatter(
                '%(message)s')

            logr_file.handlers = [handfile]

            handfile.setLevel(level)
            handfile.setFormatter(format)


        self.__started = True


    def stop(
        self,
    ) -> None:
        """
        Deinitialize the Python logging library using parameters.
        """

        logr_stdo = self.__logr_stdo
        logr_file = self.__logr_file

        logr_stdo.handlers = [NullHandler()]
        logr_file.handlers = [NullHandler()]

        self.__started = False


    def log(
        self,
        level: Literal[LOGLEVELS],
        *,
        exc_info: Optional[Exception] = None,
        **kwargs: Any,
    ) -> None:
        """
        Prepare keyword arguments and log to configured output.

        :param exc_info: Optional exception included with trace.
        :param kwargs: Keyword arguments for populating message.
        """

        if self.__started is False:
            return

        stdo_level = self.__stdo_level
        file_level = self.__file_level
        file_path = self.__file_path

        logr_stdo = self.__logr_stdo
        logr_file = self.__logr_file

        # Prevent debug from creating
        # new object each time called
        if (level == 'debug'
                and stdo_level != 'debug'
                and file_level != 'debug'):
            return None

        message = Message(level, **kwargs)

        if stdo_level is not None:

            logr_stdo.log(
                level=LOGSEVERS[level],
                msg=message.stdo_output,
                exc_info=exc_info)

        if file_path and file_level:

            logr_file.log(
                level=LOGSEVERS[level],
                msg=message.file_output,
                exc_info=exc_info)


    def log_c(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Prepare keyword arguments and log to configured output.

        :param kwargs: Keyword arguments for populating message.
        """

        self.log('critical', **kwargs)


    def log_d(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Prepare keyword arguments and log to configured output.

        :param kwargs: Keyword arguments for populating message.
        """

        self.log('debug', **kwargs)


    def log_e(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Prepare keyword arguments and log to configured output.

        :param kwargs: Keyword arguments for populating message.
        """

        self.log('error', **kwargs)


    def log_i(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Prepare keyword arguments and log to configured output.

        :param kwargs: Keyword arguments for populating message.
        """

        self.log('info', **kwargs)


    def log_w(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Prepare keyword arguments and log to configured output.

        :param kwargs: Keyword arguments for populating message.
        """

        self.log('warning', **kwargs)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from pathlib import Path

from pytest import RaisesExc
from pytest import fixture
from pytest import raises

from . import SAMPLES
from . import _DICT1R
from ..notate import delate
from ..notate import expate
from ..notate import getate
from ..notate import impate
from ..notate import setate
from ..types import DictStrAny
from ...utils import load_sample
from ...utils import prep_sample
from ...utils.sample import ENPYRWS



_SAMPLE = dict[str, DictStrAny]



@fixture
def sample_impate() -> _SAMPLE:
    """
    Construct the dictionary for use with downstream tests.

    :returns: Newly constructed dictionary for use in tests.
    """

    recurse_implode = (
        impate(
            deepcopy(_DICT1R),
            recurse_list=True,
            implode_list=True))

    assert isinstance(
        recurse_implode, dict)

    nocurse_noplode = (
        impate(
            deepcopy(_DICT1R),
            recurse_list=False,
            implode_list=False))

    assert isinstance(
        nocurse_noplode, dict)

    recurse_noplode = (
        impate(
            deepcopy(_DICT1R),
            recurse_list=True,
            implode_list=False))

    assert isinstance(
        recurse_noplode, dict)

    nocurse_implode = (
        impate(
            deepcopy(_DICT1R),
            recurse_list=False,
            implode_list=True))

    assert isinstance(
        nocurse_implode, dict)


    return {
        'recurse_implode': recurse_implode,
        'nocurse_noplode': nocurse_noplode,
        'recurse_noplode': recurse_noplode,
        'nocurse_implode': nocurse_implode}



@fixture
def sample_expate(
    sample_impate: _SAMPLE,
) -> _SAMPLE:
    """
    Construct the dictionary for use with downstream tests.

    :param sample_impate: Source dictionary for use in test.
    :returns: Newly constructed dictionary for use in tests.
    """

    recurse_implode = (
        sample_impate[
            'recurse_implode'])

    assert isinstance(
        recurse_implode, dict)

    nocurse_noplode = (
        sample_impate[
            'nocurse_noplode'])

    assert isinstance(
        nocurse_noplode, dict)

    recurse_noplode = (
        sample_impate[
            'recurse_noplode'])

    assert isinstance(
        recurse_noplode, dict)

    nocurse_implode = (
        sample_impate[
            'nocurse_implode'])

    assert isinstance(
        nocurse_implode, dict)

    return {
        'recurse_implode': (
            expate(recurse_implode)),
        'nocurse_noplode': (
            expate(nocurse_noplode)),
        'recurse_noplode': (
            expate(recurse_noplode)),
        'nocurse_implode': (
            expate(nocurse_implode))}



def test_getate() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    source = deepcopy(_DICT1R)


    value = getate(['1', 2], '1')
    assert value == 2

    value = getate((1, 2), '1')
    assert value == 2

    value = getate({'1': 2}, '1')
    assert value == 2


    path = 'recurse/dict/key'
    value = getate(source, path)

    assert value == 'd1dict'


    path = 'recurse/list/0'
    value = getate(source, path)

    assert value == 'd1list'



def test_getate_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    source = deepcopy(_DICT1R)


    assert not getate({}, 'd/n/e')
    assert not getate([], '0/n/e')


    path = 'recurse/str/a'
    value = getate(source, path)

    assert value is None



def test_setate() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    source = deepcopy(_DICT1R)


    path = 'list/1'
    before = getate(source, path)
    setate(source, path, 1)
    after = getate(source, path)
    assert after == 1
    assert before is None


    path = 'recurse/dict/key'
    before = getate(source, path)
    setate(source, path, 1)
    after = getate(source, path)
    assert after == 1
    assert before == 'd1dict'


    path = 'nested/0/dict/key'
    before = getate(source, path)
    setate(source, path, 1)
    after = getate(source, path)
    assert after == 1
    assert before == 'd1dict'


    path = 'recurse/list/0'
    before = getate(source, path)
    setate(source, path, 1)
    after = getate(source, path)
    assert after == 1
    assert before == 'd1list'



def test_setate_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    source = deepcopy(_DICT1R)


    path = 'nested/1/dict/key'
    before = getate(source, path)
    setate(source, path, 1)
    after = getate(source, path)
    assert after == 1
    assert before is None



def test_setate_raises() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    _raises: RaisesExc[
        ValueError | IndexError]


    _raises = raises(ValueError)

    with _raises as reason:
        setate(1, '1', 1)  # type: ignore

    _reason = str(reason.value)

    assert _reason == 'source'


    _raises = raises(IndexError)

    with _raises as reason:
        setate([], '1', 1)

    _reason = str(reason.value)

    assert _reason == '1'



def test_delate() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    source = deepcopy(_DICT1R)


    path = 'recurse/dict/key'
    before = getate(source, path)
    delate(source, path)
    after = getate(source, path)
    assert after is None
    assert before == 'd1dict'


    path = 'nested/0/dict/key'
    before = getate(source, path)
    delate(source, path)
    after = getate(source, path)
    assert after is None
    assert before == 'd1dict'


    path = 'recurse/list/0'
    before = getate(source, path)
    delate(source, path)
    after = getate(source, path)
    assert after is None
    assert before == 'd1list'



def test_delate_raises() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    _raises = raises(ValueError)

    with _raises as reason:
        delate(1, '1')  # type: ignore

    _reason = str(reason.value)

    assert _reason == 'source'


    _raises = raises(ValueError)

    with _raises as reason:
        delate({'a': 1}, 'a/1/c')

    _reason = str(reason.value)

    assert _reason == 'source'



def test_impate(
    tmp_path: Path,
    sample_impate: _SAMPLE,
    sample_expate: _SAMPLE,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param sample_impate: Source dictionary for use in test.
    :param sample_expate: Source dictionary for use in test.
    :param tmp_path: pytest object for temporal filesystem.
    """

    sample_path = (
        SAMPLES / 'impate.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=sample_impate)

    expect = prep_sample(
        content=sample_impate)

    assert expect == sample



def test_impate_cover(
    tmp_path: Path,
    sample_impate: _SAMPLE,
    sample_expate: _SAMPLE,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param sample_impate: Source dictionary for use in test.
    :param sample_expate: Source dictionary for use in test.
    :param tmp_path: pytest object for temporal filesystem.
    """


    default = sample_impate[
        'recurse_implode']

    assert default == impate(default)


    sample = 'recurse_noplode'

    implode = impate(
        [sample_expate[sample],
         sample_expate[sample]],
        implode_list=False,
        recurse_list=True)

    assert isinstance(implode, list)

    assert implode[0] == (
        sample_impate[sample])


    source = {'foo': {'bar': 'baz'}}
    expect = {'foo/bar': 'baz'}

    assert impate(source) == expect



def test_impate_raises() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    _raises: RaisesExc[ValueError]


    _raises = raises(ValueError)

    with _raises as reason:
        impate('foo')  # type: ignore

    _reason = str(reason.value)

    assert _reason == 'source'



def test_expate(
    tmp_path: Path,
    sample_impate: _SAMPLE,
    sample_expate: _SAMPLE,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param sample_impate: Source dictionary for use in test.
    :param sample_expate: Source dictionary for use in test.
    :param tmp_path: pytest object for temporal filesystem.
    """

    sample_path = (
        SAMPLES / 'expate.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=sample_expate)

    expect = prep_sample(
        content=sample_expate)

    assert expect == sample



def test_expate_cover(
    tmp_path: Path,
    sample_impate: _SAMPLE,
    sample_expate: _SAMPLE,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param sample_impate: Source dictionary for use in test.
    :param sample_expate: Source dictionary for use in test.
    :param tmp_path: pytest object for temporal filesystem.
    """


    assert sample_expate == (
        expate(sample_impate))


    assert sample_expate == (
        expate(sample_expate))


    assert all(
        x == _DICT1R for x in
        sample_expate.values())



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..classes import clsname
from ..classes import lattrs
from ...config import LoggerParams

if TYPE_CHECKING:
    from ...config import Config



def test_BaseModel() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    params = LoggerParams(
        stdo_level='info')

    dumped = params.endumped
    pruned = params.enpruned

    assert dumped == {
        'file_level': None,
        'file_path': None,
        'stdo_level': 'info'}

    assert pruned == {
        'stdo_level': 'info'}



def test_clsname(
    config: 'Config',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param config: Primary class instance for configuration.
    """

    assert clsname(config) == 'Config'



def test_lattrs(
    config: 'Config',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param config: Primary class instance for configuration.
    """

    attrs = lattrs(config)

    assert attrs == [
        '_Config__model',
        '_Config__files',
        '_Config__cargs',
        '_Config__sargs',
        '_Config__params',
        '_Config__paths',
        '_Config__logger',
        '_Config__crypts',
        '_Config__jinja2']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import copy
from copy import deepcopy

from ..classes import lattrs
from ..empty import Empty
from ..empty import EmptyType



def test_EmptyType() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    empty = EmptyType()


    attrs = lattrs(empty)

    assert attrs == [
        '_EmptyType__empty']


    assert repr(empty) == 'Empty'

    assert isinstance(
        hash(empty), int)

    assert str(empty) == 'Empty'


    assert not (Empty or None)
    assert Empty is empty
    assert Empty is Empty
    assert Empty is EmptyType()
    assert Empty == empty
    assert Empty == Empty
    assert Empty == EmptyType()
    assert Empty is not None
    assert Empty != 'Empty'

    assert deepcopy(Empty) is Empty
    assert copy(Empty) is Empty



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..strings import hasstr
from ..strings import inrepr
from ..strings import instr
from ..strings import rplstr
from ..strings import strplwr



def test_strplwr() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert strplwr(' Foo ') == 'foo'



def test_hasstr() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    assert hasstr('abc', 'a')
    assert hasstr('abc', 'b')
    assert hasstr('abc', 'c')



def test_inrepr() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    class MyClass:
        pass

    item = MyClass()

    assert inrepr('MyClass', item)



def test_instr() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    class MyClass:
        pass

    item = MyClass()

    assert instr('MyClass', item)



def test_rplstr() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    string = rplstr('foo', 'o', 'O')

    assert string == 'fOO'



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from pytest import fixture

from . import SAMPLES
from ..paths import ConfigPath
from ..paths import ConfigPaths
from ... import PROJECT
from ...types import inrepr
from ...types import instr
from ...types import lattrs
from ...utils import load_sample
from ...utils import prep_sample
from ...utils.sample import ENPYRWS



@fixture
def paths(
    config_path: Path,
) -> ConfigPaths:
    """
    Construct the instance for use in the downstream tests.

    :param config_path: Custom fixture for populating paths.
    :returns: Newly constructed instance of related class.
    """

    return ConfigPaths([
        f'{SAMPLES}/stark',
        f'{SAMPLES}/wayne'])



def test_ConfigPath(
    config_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param config_path: Custom fixture for populating paths.
    """

    path = ConfigPath(config_path)


    attrs = lattrs(path)

    assert attrs == [
        'path',
        'config']


    assert inrepr(
        'paths.ConfigPath object',
        path)

    assert isinstance(
        hash(path), int)

    assert instr(
        'paths.ConfigPath object',
        path)


    assert 'test' in path.path.name

    assert len(path.config) == 1



def test_ConfigPaths(
    paths: ConfigPaths,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param paths: Custom fixture for the configuration paths.
    """


    attrs = lattrs(paths)

    assert attrs == [
        'paths', 'config',
        '_ConfigPaths__merge']


    assert inrepr(
        'paths.ConfigPaths object',
        paths)

    assert isinstance(
        hash(paths), int)

    assert instr(
        'paths.ConfigPaths object',
        paths)


    assert len(paths.paths) == 2

    assert len(paths.config) == 2


    replaces = {
        'PROJECT': PROJECT}

    sample_path = (
        SAMPLES / 'paths.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=paths.merge,
        replace=replaces)

    expect = prep_sample(
        content=paths.merge,
        replace=replaces)

    assert expect == sample



def test_ConfigPaths_cover(
    paths: ConfigPaths,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param paths: Custom fixture for the configuration paths.
    """

    merge1 = paths.merge
    merge2 = paths.merge

    assert merge1 is not merge2



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .jinja2 import Jinja2
from .network import Network
from .network import insubnet_ip
from .network import isvalid_ip


__all__ = [
    'Jinja2',
    'Network',
    'insubnet_ip',
    'isvalid_ip']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from _pytest.capture import CaptureFixture

from ..stdout import ArrayColors
from ..stdout import array_ansi
from ..stdout import kvpair_ansi
from ..stdout import make_ansi
from ..stdout import print_ansi
from ..stdout import strip_ansi
from ...config import LoggerParams
from ...times import Duration
from ...times import Time
from ...times.common import UNIXMPOCH
from ...types import Empty



def test_print_ansi(
    capsys: CaptureFixture[str],
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param capsys: pytest object for capturing print message.
    """

    print_ansi('<c91>test<c0>', 'print')

    output = capsys.readouterr().out

    assert strip_ansi(output) == 'test\n'



def test_make_ansi() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    output = make_ansi('<c31>test<c0>')

    assert output == (
        '\x1b[0;31mtest\x1b[0;0m')



def test_kvpair_ansi() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    output = kvpair_ansi('key', 'value')

    assert output == (
        '\x1b[0;90mkey\x1b[0;37m="\x1b'
        '[0;0mvalue\x1b[0;37m"\x1b[0;0m')

    output = kvpair_ansi('key', [1, 2])

    assert output == (
        '\x1b[0;90mkey\x1b[0;37m="\x1b'
        '[0;0m1,2\x1b[0;37m"\x1b[0;0m')

    output = kvpair_ansi('key', None)

    assert output == (
        '\x1b[0;90mkey\x1b[0;37m="\x1b'
        '[0;0mNone\x1b[0;37m"\x1b[0;0m')



def test_strip_ansi() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    output = '\x1b[0;31mtest\x1b[0;0m'

    assert strip_ansi(output) == 'test'



def test_array_ansi() -> None:  # noqa: CFQ001
    """
    Perform various tests associated with relevant routines.
    """

    simple = {
        'str': 'value',
        'list': [1, 2],
        'bool': False}


    output = strip_ansi(
        array_ansi(simple))

    assert output == (
        "str: 'value'\n"
        'list: list\n'
        '  - 1\n'
        '  - 2\n'
        'bool: False')


    output = strip_ansi(
        array_ansi([simple]))

    assert output == (
        '- dict\n'
        "  str: 'value'\n"
        '  list: list\n'
        '    - 1\n'
        '    - 2\n'
        '  bool: False')


    output = strip_ansi(
        array_ansi((1, 2, 3)))

    assert output == (
        '- 1\n'
        '- 2\n'
        '- 3')


    output = strip_ansi(
        array_ansi({1, 2, 3}))

    assert output == (
        '- 1\n'
        '- 2\n'
        '- 3')


    colors = ArrayColors()

    params = LoggerParams()

    repeat = {
        'dict': simple | {
            'dict': simple}}

    durate = Duration(190802)

    source = {
        'str': 'value',
        'int': 1,
        'float': 1.0,
        'complex': complex(3, 1),
        'list': [simple],
        'tuple': (simple,),
        'range': range(1, 3),
        'dict1': simple,
        'dict2': simple,
        'dict3': simple,
        'set': {1, 2, 3},
        'frozenset': {1, 2, 3},
        'bool': True,
        'none': None,
        '_private': None,
        'repeat': repeat,
        'colors': colors,
        'params': params,
        'Empty': Empty,
        'Time': Time(0),
        'Duration': durate}


    output = strip_ansi(
        array_ansi(source))

    assert output == (
        "str: 'value'\n"
        'int: 1\n'
        'float: 1.0\n'
        'complex: (3+1j)\n'
        'list: list\n'
        '  - dict\n'
        "    str: 'value'\n"
        '    list: list\n'
        '      - 1\n'
        '      - 2\n'
        '    bool: False\n'
        'tuple: tuple\n'
        '  - dict\n'
        "    str: 'value'\n"
        '    list: list\n'
        '      - 1\n'
        '      - 2\n'
        '    bool: False\n'
        'range: range(1, 3)\n'
        'dict1: dict\n'
        "  str: 'value'\n"
        '  list: list\n'
        '    - 1\n'
        '    - 2\n'
        '  bool: False\n'
        'dict2: dict\n'
        "  str: 'value'\n"
        '  list: list\n'
        '    - 1\n'
        '    - 2\n'
        '  bool: False\n'
        'dict3: dict\n'
        "  str: 'value'\n"
        '  list: list\n'
        '    - 1\n'
        '    - 2\n'
        '  bool: False\n'
        'set: set\n'
        '  - 1\n'
        '  - 2\n'
        '  - 3\n'
        'frozenset: set\n'
        '  - 1\n'
        '  - 2\n'
        '  - 3\n'
        'bool: True\n'
        'none: None\n'
        '_private: None\n'
        'repeat: dict\n'
        '  dict: dict\n'
        "    str: 'value'\n"
        '    list: list\n'
        '      - 1\n'
        '      - 2\n'
        '    bool: False\n'
        '    dict: dict\n'
        "      str: 'value'\n"
        '      list: REPEAT\n'
        '      bool: False\n'
        f'colors: ArrayColors\n'
        '  label: 37\n'
        '  key: 97\n'
        '  colon: 37\n'
        '  hyphen: 37\n'
        '  bool: 93\n'
        '  none: 33\n'
        '  str: 92\n'
        '  num: 93\n'
        '  times: 96\n'
        '  empty: 36\n'
        '  other: 91\n'
        'params: LoggerParams\n'
        '  stdo_level: None\n'
        '  file_level: None\n'
        '  file_path: None\n'
        'Empty: Empty\n'
        f'Time: {UNIXMPOCH}\n'
        'Duration: 2d5h')



def test_array_ansi_cover() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    colors = ArrayColors()

    output = strip_ansi(
        array_ansi(colors))

    assert output == (
        'label: 37\n'
        'key: 97\n'
        'colon: 37\n'
        'hyphen: 37\n'
        'bool: 93\n'
        'none: 33\n'
        'str: 92\n'
        'num: 93\n'
        'times: 96\n'
        'empty: 36\n'
        'other: 91')


    params = LoggerParams()

    output = strip_ansi(
        array_ansi(params))

    assert output == (
        'stdo_level: None\n'
        'file_level: None\n'
        'file_path: None')



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from inspect import currentframe



def funcname() -> str:
    """
    Return the current function name where code is running.

    :returns: Current function name where code is running.
    """

    frame = currentframe()

    assert frame is not None, (
        'Frame not present')

    caller = frame.f_back

    assert caller is not None, (
        'Caller not present')

    name = caller.f_code.co_name
    focals = caller.f_locals


    if 'self' in focals:

        parent = (
            focals['self']
            .__class__
            .__name__)

        return f'{parent}.{name}'

    elif 'cls' in focals:

        parent = (
            focals['cls']
            .__name__)

        return f'{parent}.{name}'


    return name



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from _pytest.logging import LogCaptureFixture

from pytest import fixture

from ..logger import Logger
from ..logger import Message
from ..params import LoggerParams
from ...times import Time
from ...times.common import UNIXMPOCH
from ...times.common import UNIXSPOCH
from ...types import inrepr
from ...types import instr
from ...types import lattrs
from ...utils import strip_ansi



_CAPLOG = list[tuple[str, int, str]]



@fixture
def logger(
    tmp_path: Path,
) -> Logger:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    params = LoggerParams(
        stdo_level='debug',
        file_level='info',
        file_path=f'{tmp_path}/test.log')

    return Logger(params)



def test_Message() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    message = Message(
        time=UNIXMPOCH,
        level='info',
        string='foobar',
        none=None,
        list=['1', 2],
        tuple=('1', 2),
        set={'1', 2},
        dict={'1': 2},
        empty_list=[],
        empty_dict={},
        int=1,
        float=2.0,
        elapsed=3.69420)


    attrs = lattrs(message)

    assert attrs == [
        '_Message__level',
        '_Message__time',
        '_Message__fields']


    assert inrepr(
        'Message(level="info"',
        message)

    assert isinstance(
        hash(message), int)

    assert instr(
        'Message(level="info"',
        message)


    assert message.level == 'info'

    assert message.time == 0

    assert len(message.fields) == 8


    output = strip_ansi(
        message.stdo_output)

    assert output == (
        'level="info"'
        f' time="{UNIXSPOCH}"'
        ' string="foobar"'
        ' list="1,2"'
        ' tuple="1,2"'
        ' set="1,2"'
        ' dict="{\'1\': 2}"'
        ' int="1"'
        ' float="2.0"'
        ' elapsed="3.69"')


    output = message.file_output

    assert output == (
        '{"level": "info",'
        f' "time": "{UNIXMPOCH}",'
        ' "string": "foobar",'
        ' "list": "1,2",'
        ' "tuple": "1,2",'
        ' "set": "1,2",'
        ' "dict": "{\'1\': 2}",'
        ' "int": "1",'
        ' "float": "2.0",'
        ' "elapsed": "3.69"}')



def test_Logger(
    logger: Logger,
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param logger: Custom fixture for the logger instance.
    :param caplog: pytest object for capturing log message.
    """


    attrs = lattrs(logger)

    assert attrs == [
        '_Logger__params',
        '_Logger__stdo_level',
        '_Logger__file_level',
        '_Logger__file_path',
        '_Logger__started',
        '_Logger__logr_stdo',
        '_Logger__logr_file']


    assert inrepr(
        'logger.Logger object',
        logger)

    assert isinstance(
        hash(logger), int)

    assert instr(
        'logger.Logger object',
        logger)


    assert logger.params

    assert logger.stdo_level == 'debug'

    assert logger.file_level == 'info'

    assert logger.file_path

    assert not logger.started

    assert logger.logger_stdo

    assert logger.logger_file


    logger.log_d(msg='pytest')
    logger.log_c(msg='pytest')
    logger.log_e(msg='pytest')
    logger.log_i(msg='pytest')
    logger.log_w(msg='pytest')



def test_Logger_cover(
    logger: Logger,
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param logger: Custom fixture for the logger instance.
    :param caplog: pytest object for capturing log message.
    """

    time = Time('now')


    # Test preventing debug from
    # any needless instantiation
    setattr(
        logger,
        '_Logger__file_level',
        'info')

    # Test preventing debug from
    # any needless instantiation
    setattr(
        logger,
        '_Logger__stdo_level',
        'info')


    def _logger_logs() -> None:
        logger.log_d(msg='pytest')
        logger.log_c(msg='pytest')
        logger.log_e(msg='pytest')
        logger.log_i(msg='pytest')
        logger.log_w(msg='pytest')
        logger.log_i(elapsed=time)


    def _logger_stdo() -> _CAPLOG:

        output = caplog.record_tuples
        key = 'encommon.logger.stdo'

        return [
            x for x in output
            if x[0] == key]


    def _logger_file() -> _CAPLOG:

        output = caplog.record_tuples
        key = 'encommon.logger.file'

        return [
            x for x in output
            if x[0] == key]


    logger.start()

    _logger_logs()

    assert len(_logger_stdo()) == 5
    assert len(_logger_file()) == 5

    logger.stop()


    _logger_logs()

    assert len(_logger_stdo()) == 5
    assert len(_logger_file()) == 5


    logger.start()

    message = 'unknown exception'
    raises = Exception('pytest')

    logger.log_e(
        message=message,
        exc_info=raises)

    stdo = _logger_stdo()[-1][2]
    file = _logger_file()[-1][2]

    assert message in str(stdo)
    assert message in str(file)

    assert len(_logger_stdo()) == 6
    assert len(_logger_file()) == 6


    logger.stop()


    _logger_logs()

    assert len(_logger_stdo()) == 6
    assert len(_logger_file()) == 6



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Union
from typing import get_args

from ..types.strings import COMMAS
from ..types.strings import SEMPTY
from ..types.strings import SPACED



DURAGROUP = Literal[
    'year', 'month', 'week',
    'day', 'hour', 'minute',
    'second']

DURASHORT = Literal[
    'y', 'mon', 'w',
    'd', 'h', 'm', 's']

DURAMAPS = dict(zip(
    get_args(DURAGROUP),
    get_args(DURASHORT)))

DURAUNITS = Union[
    DURAGROUP, DURASHORT]



class Duration:
    """
    Convert the provided seconds in a human friendly format.

    Example: Common Methods
    -----------------------
    >>> durate = Duration(6048e4)
    >>> durate.short
    '1y 11mon 5d'
    >>> durate.compact
    '1y11mon5d'
    >>> durate.verbose
    '1 year, 11 months, 5 days'

    Example: Dump the Units
    -----------------------
    >>> durate = Duration(6048e4)
    >>> durate.units()
    {'year': 1, 'month': 11, 'day': 5}

    Example: Disable Smart Seconds
    ------------------------------
    >>> durate = Duration(7201, False)
    >>> durate.compact
    '2h1s'
    >>> durate.verbose
    '2 hours, 1 second'

    Example: Basic Operators
    ------------------------
    >>> durate1 = Duration(1000)
    >>> durate2 = Duration(2000)
    >>> durate1 == durate2
    False
    >>> durate1 != durate2
    True
    >>> durate1 > durate2
    False
    >>> durate1 >= durate2
    False
    >>> durate1 < durate2
    True
    >>> durate1 <= durate2
    True
    >>> durate1 + durate2
    3000.0
    >>> durate1 - durate2
    -1000.0
    >>> durate2 - durate1
    1000.0

    :param seconds: Period in seconds that will be iterated.
    :param smart: Determines if we hide seconds after minute.
    :param groups: Determine the quantity of groups to show,
        ensuring larger units are returned before smaller.
    """

    __source: float
    __smart: bool
    __groups: int


    def __init__(
        self,
        seconds: int | float,
        smart: bool = True,
        groups: int = len(DURAMAPS),
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__source = float(seconds)
        self.__smart = bool(smart)
        self.__groups = int(groups)


    def __repr__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return (
            f'Duration('
            f'seconds={self.source}, '
            f'smart={self.smart}, '
            f'groups={self.groups})')


    def __hash__(
        self,
    ) -> int:
        """
        Built-in method used when performing hashing operations.

        :returns: Integer hash value for the internal reference.
        """

        return hash(self.__source)


    def __str__(
        self,
    ) -> str:
        """
        Built-in method for representing the values for instance.

        :returns: String representation for values from instance.
        """

        return self.compact


    def __int__(
        self,
    ) -> int:
        """
        Built-in method representing numeric value for instance.

        :returns: Numeric representation for value in instance.
        """

        return int(self.__source)


    def __float__(
        self,
    ) -> float:
        """
        Built-in method representing numeric value for instance.

        :returns: Numeric representation for value in instance.
        """

        return float(self.__source)


    def __add__(
        self,
        other: Union['Duration', int, float],
    ) -> float:
        """
        Built-in method for mathematically processing the value.

        :param other: Other value being compared with instance.
        :returns: Python timedelta object containing the answer.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source + other


    def __sub__(
        self,
        other: Union['Duration', int, float],
    ) -> float:
        """
        Built-in method for mathematically processing the value.

        :param other: Other value being compared with instance.
        :returns: Python timedelta object containing the answer.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source - other


    def __eq__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source == other


    def __ne__(
        self,
        other: object,
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        return not self.__eq__(other)


    def __gt__(
        self,
        other: Union['Duration', int, float],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source > other


    def __ge__(
        self,
        other: Union['Duration', int, float],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source >= other


    def __lt__(
        self,
        other: Union['Duration', int, float],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source < other


    def __le__(
        self,
        other: Union['Duration', int, float],
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        if hasattr(other, 'source'):
            other = other.source

        return self.__source <= other


    @property
    def source(
        self,
    ) -> float:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__source


    @property
    def smart(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__smart


    @property
    def groups(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__groups


    def units(
        self,
        short: bool = False,
    ) -> dict[DURAUNITS, int]:
        """
        Return the groups of time units with each relevant value.

        :param short: Determine if we should use the short hand.
        :returns: Groups of time units with each relevant value.
        """

        source = self.__source
        seconds = int(source)

        units: dict[DURAUNITS, int] = {}

        groups: dict[DURAGROUP, int] = {
            'year': 31536000,
            'month': 2592000,
            'week': 604800,
            'day': 86400,
            'hour': 3600,
            'minute': 60}


        items = groups.items()

        for key, value in items:

            if seconds < value:
                continue

            _value = seconds // value
            seconds %= value

            units[key] = _value


        if ((seconds >= 1
                and source > 60)
                or source < 60):
            units['second'] = seconds

        if ('second' in units
                and len(units) >= 2
                and self.__smart):
            del units['second']


        _groups = (
            list(units.items())
            [:self.__groups])

        if short is False:
            return dict(_groups)

        return {
            DURAMAPS[k]: v
            for k, v in _groups}


    def __duration(
        self,
        delim: str = SPACED,
        short: bool = True,
    ) -> str:
        """
        Return the compact format determined by source duration.

        :param delim: Optional delimiter for between the groups.
        :param short: Determine if we should use the short hand.
        :returns: Compact format determined by source duration.
        """

        parts: list[str] = []

        groups = self.units(short)

        space = (
            SEMPTY if short
            else SPACED)


        items = groups.items()

        for part, value in items:

            unit: str = part

            if (short is False
                    and value != 1):
                unit += 's'

            parts.append(
                f'{value}{space}{unit}')


        return delim.join(parts)


    @property
    def short(
        self,
    ) -> str:
        """
        Return the compact format determined by source duration.

        :returns: Compact format determined by source duration.
        """

        return self.__duration()


    @property
    def compact(
        self,
    ) -> str:
        """
        Return the compact format determined by source duration.

        :returns: Compact format determined by source duration.
        """

        return self.short.replace(
            SPACED, SEMPTY)


    @property
    def verbose(
        self,
    ) -> str:
        """
        Return the verbose format determined by source duration.

        :returns: Verbose format determined by source duration.
        """

        if self.__source < 60:
            return 'just now'

        return self.__duration(
            COMMAS, False)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .crypts import Crypts
from .hashes import Hashes
from .params import CryptsParams



__all__ = [
    'Crypts',
    'CryptsParams',
    'Hashes']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .duration import Duration
from .params import TimerParams
from .params import TimersParams
from .params import WindowParams
from .params import WindowsParams
from .parse import parse_time
from .parse import shift_time
from .parse import since_time
from .parse import string_time
from .time import Time
from .timer import Timer
from .timers import Timers
from .unitime import unitime
from .utils import findtz
from .window import Window
from .windows import Windows



__all__ = [
    'Duration',
    'Timer',
    'TimerParams',
    'Timers',
    'TimersParams',
    'Time',
    'Window',
    'WindowParams',
    'Windows',
    'WindowsParams',
    'findtz',
    'parse_time',
    'shift_time',
    'since_time',
    'string_time',
    'unitime']



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from datetime import datetime
from datetime import timezone
from datetime import tzinfo
from typing import Any
from typing import Optional

from dateutil.tz import gettz



def findtz(
    tzname: Optional[str] = None,
) -> tzinfo:
    """
    Return the located timezone object for the provided name.

    Example
    -------
    >>> findtz('US/Central')
    tzfile('/usr/share/zoneinfo/US/Central')

    :param tzname: Name of the timezone associated to source.
    :returns: Located timezone object for the provided name.
    """

    if tzname is None:
        return timezone.utc

    tzinfo = gettz(tzname)

    if tzinfo is None:
        raise ValueError('tzname')

    return tzinfo



def utcdatetime(
    *args: Any,
    **kwargs: Any,
) -> datetime:
    """
    Return the instance of datetime within the UTC timezone.

    .. warning::
       If no arguments are provided, returns current time.

    Example
    -------
    >>> utcdatetime(1970, 1, 1)
    datetime.datetime(1970, 1, 1, 0...

    :param args: Positional arguments passed for downstream.
    :param kwargs: Keyword arguments passed for downstream.
    :returns: Instance of datetime within the UTC timezone.
    """

    tzinfo = timezone.utc

    if not args and not kwargs:
        return datetime.now(tz=tzinfo)

    if 'tzinfo' not in kwargs:
        kwargs['tzinfo'] = tzinfo

    return (
        datetime(*args, **kwargs)
        .astimezone(timezone.utc))



def strptime(
    source: str,
    formats: str | list[str],
) -> datetime:
    """
    Parse provided time value with various supported formats.

    Example
    -------
    >>> strptime('2023', '%Y')
    datetime.datetime(2023, 1, 1, 0...

    :param source: Time in various forms that will be parsed.
    :param formats: Various formats compatable with strptime.
    :returns: Python datetime object containing related time.
    """

    tzinfo = timezone.utc

    if isinstance(formats, str):
        formats = [formats]


    def _strptime(
        format: str,
    ) -> datetime:

        return (
            datetime
            .strptime(source, format)
            .astimezone(tzinfo))


    for format in formats:

        with suppress(ValueError):
            return _strptime(format)


    raise ValueError('invalid')



def strftime(
    source: datetime,
    format: str,
) -> str:
    """
    Return the timestamp string for datetime object provided.

    .. note::
       This function is extremely pedantic and cosmetic.

    Example
    -------
    >>> dtime = datetime(2023, 1, 1)
    >>> strftime(dtime, '%Y')
    '2023'

    :param source: Python datetime instance containing source.
    :param format: Format for the timestamp string returned.
    :returns: Timestamp string for datetime object provided.
    """

    return (
        datetime
        .strftime(source, format))



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import asdict
from dataclasses import is_dataclass
from json import dumps
from os import environ
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional


from .files import read_text
from .files import save_text
from ..types import BaseModel
from ..types import DictStrAny
from ..types import rplstr



PREFIX = 'encommon_sample'

ENPYRWS = (
    environ.get('ENPYRWS') == '1')



def prep_sample(  # noqa: CFQ004
    content: Any,
    *,
    default: Callable[[Any], str] = str,
    replace: Optional[DictStrAny] = None,
    indent: Optional[int] = 2,
) -> str:
    r"""
    Return the content after processing as the sample value.

    Example
    -------
    >>> prep_sample(['one', 'two'])
    '[\n  "one",\n  "two"\n]'

    Example
    -------
    >>> from ..types import Empty
    >>> prep_sample({'one': Empty})
    '{\n  "one": "Empty"\n}'

    :param content: Content that will be processed as JSON.
    :param default: Callable used when stringifying values.
    :param replace: Optional values to replace in the file.
    :returns: Content after processing as the sample value.
    """


    def _default(
        value: Any,  # noqa: ANN401
    ) -> DictStrAny | str:

        if is_dataclass(value):

            assert not isinstance(
                value, type)

            return asdict(value)

        if isinstance(value, BaseModel):
            return value.endumped

        return str(value)


    content = dumps(
        content,
        default=_default,
        indent=indent)

    replace = replace or {}

    items = replace.items()

    for old, new in items:

        new = str(new)

        old = f'_/{PREFIX}/{old}/_'

        content = rplstr(
            content, new, old)

    return str(content)



def load_sample(
    path: str | Path,
    content: Optional[Any] = None,
    update: bool = False,
    *,
    default: Callable[[Any], str] = str,
    replace: Optional[DictStrAny] = None,
) -> str:
    r"""
    Load the sample file and compare using provided content.

    .. testsetup::
       >>> from json import dumps
       >>> from json import loads
       >>> path = Path(getfixture('tmpdir'))
       >>> sample = path.joinpath('sample')

    Example
    -------
    >>> content = {'one': 'two'}
    >>> load_sample(sample, content)
    '{\n  "one": "two"\n}'

    Example
    -------
    >>> load_sample(sample)
    '{\n  "one": "two"\n}'

    :param path: Complete or relative path for sample file.
    :param update: Determine whether the sample is updated.
    :param content: Content that will be processed as JSON.
    :param default: Callable used when stringifying values.
    :param replace: Optional values to replace in the file.
    :returns: Content after processing using JSON functions.
    """


    path = Path(path).resolve()

    loaded: Optional[Any] = None


    content = prep_sample(
        content=content,
        default=default,
        replace=replace)


    def _save_sample() -> None:
        save_text(path, content)


    def _load_sample() -> str:
        return read_text(path)


    if path.exists():
        loaded = _load_sample()

    if not path.exists():
        _save_sample()

    elif (update is True
            and content is not None
            and content != loaded):
        _save_sample()


    return _load_sample()



def read_sample(
    sample: str,
    *,
    replace: Optional[DictStrAny] = None,
    prefix: bool = True,
) -> str:
    """
    Return the content after processing as the sample value.

    :param sample: Content that will be processed as sample.
    :param replace: Optional values to replace in the file.
    :param prefix: Determine whether or not prefix is added.
    :returns: Content after processing as the sample value.
    """

    replace = replace or {}

    items = replace.items()

    for new, old in items:

        if prefix is True:
            old = f'_/{PREFIX}/{old}/_'

        sample = rplstr(
            sample, new, old)

    return str(sample)



def rvrt_sample(
    sample: str,
    *,
    replace: Optional[DictStrAny] = None,
) -> str:
    """
    Return the content after processing as the sample value.

    :param sample: Content that will be processed as sample.
    :param replace: Optional values to replace in the file.
    :returns: Content after processing as the sample value.
    """

    replace = replace or {}

    items = replace.items()

    for new, old in items:

        new = f'_/{PREFIX}/{new}/_'

        sample = rplstr(
            sample, new, old)

    return str(sample)



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ...types import inrepr
from ...types import instr
from ...types import lattrs

if TYPE_CHECKING:
    from ..content import Content



def test_Content(
    content: 'Content',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param content: Primary class instance for the content.
    """


    attrs = lattrs(content)

    assert attrs == []


    assert inrepr(
        'content.Content',
        content)

    assert isinstance(
        hash(content), int)

    assert instr(
        'content.Content',
        content)


    script = (
        content
        .scripts('default'))

    assert script is not None


    styles = (
        content
        .styles('default'))

    assert styles is not None


    image = (
        content
        .images('enasis'))

    assert image is not None



def test_Content_cover(
    content: 'Content',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param content: Primary class instance for the content.
    """


    script = (
        content
        .scripts('doesnotexist'))

    assert script is None


    styles = (
        content
        .styles('doesnotexist'))

    assert styles is None


    image = (
        content
        .images('doesnotexist'))

    assert image is None



"""
Functions and routines associated with Enasis Network Common Library.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..params import TimerParams
from ..params import WindowParams



def test_TimerParams() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    params = TimerParams(
        timer=1,
        start='1980-01-01T00:00:00Z')

    assert params.start == (
        '1980-01-01T00:00:00.000000+0000')



def test_WindowParams() -> None:
    """
    Perform various tests associated with relevant routines.
    """


    params = WindowParams(
        window='* * * * *',
        start='1980-01-01T00:00:00Z',
        stop='1980-01-02T00:00:00Z',
        anchor='1980-01-01T00:00:00Z',
        delay=10)

    assert params.start == (
        '1980-01-01T00:00:00.000000+0000')

    assert params.stop == (
        '1980-01-02T00:00:00.000000+0000')

    assert str(params.delay) == '10.0'


    params = WindowParams(
        window=60,
        start='1980-01-01T00:00:00Z',
        stop='1980-01-02T00:00:00Z',
        anchor='1980-01-01T00:00:00Z',
        delay=10)

    assert params.start == (
        '1980-01-01T00:00:00.000000+0000')

    assert params.stop == (
        '1980-01-02T00:00:00.000000+0000')

    assert str(params.delay) == '10.0'
